# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Helper for creating ONNX nodes
from onnx import helper as oh

# QONNX arbitrary precision data types
from qonnx.core.datatype import DataType

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper

# QONNX graph transformation base class
from qonnx.transformation.base import Transformation

# QONNX graph transformations for inferring datatypes and shapes
from qonnx.transformation.infer_shapes import InferShapes

# Gets items from protobuf by name
from qonnx.util.basic import get_by_name


class CollapseRepeatedOp(Transformation):
    """Collapse repeated consecutive operations with constant parameters into
    a single operation. make_collapsed_param_fxn must take two tensors and
    return a tensor which gives the equivalent result using a single op."""

    def __init__(self, op_name, make_collapsed_param_fxn):
        super().__init__()
        self.op_name = op_name
        self.make_collapsed_param_fxn = make_collapsed_param_fxn

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        for n in graph.node:
            node_ind += 1
            if (
                n.op_type == self.op_name
                and not model.is_fork_node(n)
                and not model.is_join_node(n)
            ):
                consumer = model.find_consumer(n.output[0])
                if (
                    consumer is not None
                    and consumer.op_type == self.op_name
                    and not model.is_join_node(consumer)
                ):
                    op0_param_name = n.input[1]
                    op1_param_name = consumer.input[1]
                    op0_param = model.get_initializer(op0_param_name)
                    op1_param = model.get_initializer(op1_param_name)
                    assert (
                        op0_param is not None
                    ), """Initializer for parameters for
                    op0 is not set."""
                    assert (
                        op1_param is not None
                    ), """Initializer for parameters for
                    op1 is not set."""
                    start_name = n.input[0]
                    end_name = consumer.output[0]
                    # compute the new parameter
                    new_param = self.make_collapsed_param_fxn(op0_param, op1_param)
                    # make and insert new node
                    new_node_param_name = op0_param_name
                    new_node = oh.make_node(
                        self.op_name, [start_name, new_node_param_name], [end_name]
                    )
                    graph.node.insert(node_ind, new_node)
                    # replace parameter value
                    model.set_initializer(new_node_param_name, new_param)
                    # be conservative with param/output DataTypes
                    model.set_tensor_datatype(new_node_param_name, DataType["FLOAT32"])
                    model.set_tensor_datatype(end_name, DataType["FLOAT32"])
                    # remove old nodes
                    graph.node.remove(n)
                    graph.node.remove(consumer)
                    graph_modified = True
        model = model.transform(InferShapes())
        return (model, graph_modified)


class CollapseRepeatedAdd(CollapseRepeatedOp):
    """Collapse repeated adder node into a single operation."""

    def __init__(self):
        super().__init__("Add", lambda x, y: y + x)


class CollapseRepeatedMul(CollapseRepeatedOp):
    """Collapse repeated multiplier node into a single operation."""

    def __init__(self):
        super().__init__("Mul", lambda x, y: y * x)


# Collapses repeated transpose operations into a single transpose operation
# having the same effect
class CollapseRepeatedTranspose(Transformation):
    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper):  # noqa
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Keep track of whether the graph has been modified
        graph_modified = False
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # Applies to Transpose operation types
            if node.op_type == "Transpose":
                # Currently does not handle fork- or join-nodes
                if model.is_fork_node(node) or model.is_join_node(node):
                    # Softly skip this node
                    continue
                # As this is not a fork-node, there can be at most one successor
                successor = model.find_direct_successors(node)
                # If Transpose is the final operation in the graph, there might
                # be no successor
                if successor is None:
                    # Softly skip this node
                    continue
                # Now there is exactly one successor which needs to be extracted
                # from the list
                successor = successor[0]
                # Successor must be a Transpose to be collapsed
                if successor.op_type != "Transpose":
                    # Softly skip this node
                    continue
                # Get the (optional) permutation indices of the first transpose
                # in case it is a multi-axis transpose
                perm1 = get_by_name(node.attribute, "perm")
                # Convert permutation indices to list of integers
                perm1 = perm1.ints if perm1 is not None else None

                # Get the (optional) permutation indices of the second transpose
                # in case it is a multi-axis transpose
                perm2 = get_by_name(successor.attribute, "perm")
                # Convert permutation indices to list of integers
                perm2 = perm2.ints if perm2 is not None else None

                # Get the shape of the input tensor
                shape = model.get_tensor_shape(
                    # fmt: off
                    node.input[0], fix_missing_init_shape=True
                    # fmt: on
                )
                # List of dimension indices in order
                dims = range(len(shape))

                # Substitute the permutation indices by the reversed index list
                # if they are not given: This is default behavior, see the docs:
                #   https://onnx.ai/onnx/operators/onnx__Transpose.html
                perm1 = list(reversed(dims)) if perm1 is None else perm1
                perm2 = list(reversed(dims)) if perm2 is None else perm2

                # Combined permutation permutes the first permutation of the
                # dimensions according to the second permutation
                perm = [perm1[i] for i in perm2]

                # Create a new Transpose operator replacing the other two
                transpose = oh.make_node(
                    # Name of the operator type
                    "Transpose",
                    # Connect to the inputs to the first transpose
                    inputs=node.input,
                    # Connect to the outputs of the second transpose
                    outputs=successor.output,
                    # Insert the new permutation indices
                    perm=perm,
                )
                # Insert the collapsed transpose operator
                graph.node.insert(index + 2, transpose)
                # Remove the two original transpose operators
                graph.node.remove(node)
                graph.node.remove(successor)
                # Track whether the graph has been modified, never resets to
                # False
                graph_modified = True
                # Break the loop after adding and removing nodes to start over
                # with a clean index
                break
        # Need to redo the shape inference after potentially removing nodes
        model = model.transform(InferShapes())  # noqa: Shadows model
        # Return the transformed model and indicate whether the graph actually
        # has been transformed
        return model, graph_modified



# Copyright (C) 2020, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import qonnx.custom_op.registry as registry
from qonnx.transformation.base import Transformation

from finn.util.fpgadataflow import is_hls_node, is_rtl_node


class ReplaceVerilogRelPaths(Transformation):
    """Convert ./ relative file paths to absolute ones for generated Verilog"""

    def __init__(self):
        super().__init__()

    def apply(self, model):
        for node in model.graph.node:
            if is_hls_node(node) or is_rtl_node(node):
                try:
                    # lookup op_type in registry of CustomOps
                    inst = registry.getCustomOp(node)
                    # find the IP gen dir
                    ipgen_path = inst.get_nodeattr("ipgen_path")
                    if ipgen_path is not None and os.path.isdir(ipgen_path):
                        for dname, dirs, files in os.walk(ipgen_path):
                            for fname in files:
                                if fname.endswith(".v"):
                                    fpath = os.path.join(dname, fname)
                                    with open(fpath, "r") as f:
                                        s = f.read()
                                    old = '$readmemh(".'
                                    new = '$readmemh("%s' % dname
                                    s = s.replace(old, new)
                                    old = '"./'
                                    new = '"%s/' % dname
                                    s = s.replace(old, new)
                                    with open(fpath, "w") as f:
                                        f.write(s)
                except KeyError:
                    pass
        return (model, False)



import argparse
import json
import matplotlib.pyplot as plt
import os
import sys
import time
from driver_instrumentation import FINNInstrumentationOverlay
from pynq import PL
from pynq.pl_server.device import Device


class FINNLiveFIFOOverlay(FINNInstrumentationOverlay):
    def __init__(
        self,
        bitfile_name,
        platform="zynq",
        fclk_mhz=100.0,
        device=None,
        download=True,
        seed=1,
        fifo_widths=dict(),
    ):
        super().__init__(
            bitfile_name,
            platform=platform,
            fclk_mhz=fclk_mhz,
            seed=seed,
            download=download,
            device=device,
        )

        self.error = False
        self.fifo_widths = fifo_widths
        self.num_fifos = len(self.fifo_widths)
        # Account for additional FIFO depth and implicit registers introduced by the virtual FIFO
        # HLS implementation that are not present in real FIFOs. This results in a minimum possible
        # FIFO depth of 1 + 8 = 9, which should be improved in a future implementation (TODO).
        self.fifo_depth_offset = 8

        # Sanity check
        # We expect 3 AXI-Lite peripherals next to the virtual FIFOs:
        # instrumentation_wrap_0, axi_gpio_0 (for reset), zynq_ps
        # We expect no additional FINN SDPs with AXI-Lite, such as runtime-writable weights
        if (len(self.ip_dict.keys()) - 3) != self.num_fifos:
            print(
                "Error: # of expected FIFOs (%d) doesn't match # of AXI-Lite interfaces (%d)"
                % (self.num_fifos, len(self.ip_dict.keys()) - 3)
            )
            self.error = True

    def configure_fifo(self, i, mode, depth=2):
        # Virtual FIFO register map
        mode_offset = 0x10
        depth_offset = 0x18
        # occupancy_offset = 0x20
        # occupancy_ctrl_offset = 0x24
        # max_occupancy_offset = 0x30
        # max_occupancy_ctrl_offset = 0x34

        ip_name = "StreamingDataflowPartition_%d" % i
        getattr(self, ip_name).write(offset=mode_offset, value=mode)
        getattr(self, ip_name).write(offset=depth_offset, value=depth)

    def total_fifo_size(self, depths):
        # Assuming FIFO SDP/AXI-Lite interfaces are ordered consistently with FIFO IDs
        total_size_bits = 0
        for i, depth in enumerate(depths):
            total_size_bits += (depth + self.fifo_depth_offset) * self.fifo_widths[str(i)]
        total_size_kB = total_size_bits / 8.0 / 1000.0
        return total_size_kB

    def size_iteratively(self, start_depth, iteration_runtime, reduction_factor=0.5):
        # Iterative FIFO-sizing function
        fifo_minimum_reached = [False] * self.num_fifos

        if isinstance(start_depth, list):
            # Individual start depth for each FIFO has been supplied
            fifo_depths = start_depth
        else:
            # Initialize all depths to the same start depth
            fifo_depths = [start_depth] * self.num_fifos

        # Reset accelerator and configure FIFOs
        self.reset_accelerator()
        for i in range(0, self.num_fifos):
            self.configure_fifo(i, mode=1, depth=fifo_depths[i])

        # Run once to determine target interval
        self.start_accelerator()
        time.sleep(1)
        (
            overflow_err,
            underflow_err,
            frame,
            checksum,
            min_latency,
            latency,
            interval,
        ) = self.observe_instrumentation(False)
        log_total_fifo_size = [int(self.total_fifo_size(fifo_depths))]
        log_interval = [interval]
        log_min_latency = [min_latency]
        log_latency = [latency]
        target_interval = interval

        # Iteratively reduce FIFO depth until all FIFOs are minimized
        iteration = 0
        start_time = time.time()
        while not all(fifo_minimum_reached):
            for fifo_id in range(0, self.num_fifos):
                if not fifo_minimum_reached[fifo_id]:
                    fifo_depth_before = fifo_depths[fifo_id]
                    fifo_depths[fifo_id] = int(fifo_depths[fifo_id] * reduction_factor)

                    # Reset accelerator
                    self.reset_accelerator()

                    # Configure all FIFOs
                    for i in range(0, self.num_fifos):
                        self.configure_fifo(i, mode=1, depth=fifo_depths[i])

                    # Start accelerator
                    self.start_accelerator()

                    # Let it run
                    time.sleep(iteration_runtime)

                    # Check if throughput dropped or deadlock occured
                    (
                        overflow_err,
                        underflow_err,
                        frame,
                        checksum,
                        min_latency,
                        latency,
                        interval,
                    ) = self.observe_instrumentation(False)

                    if interval > target_interval or interval == 0 or overflow_err or underflow_err:
                        # Revert depth reduction and mark FIFO as minimized
                        fifo_depths[fifo_id] = fifo_depth_before
                        fifo_minimum_reached[fifo_id] = True
                    else:
                        log_total_fifo_size.append(int(self.total_fifo_size(fifo_depths)))
                        log_interval.append(interval)
                        log_min_latency.append(min_latency)
                        log_latency.append(latency)

                    if fifo_depths[fifo_id] == 1:
                        fifo_minimum_reached[fifo_id] = True

            # Report status
            print("Iteration: %d" % iteration)
            print("Numer of minimized FIFOs: %d/%d" % (sum(fifo_minimum_reached), self.num_fifos))
            print("Interval: %d" % log_interval[-1])
            print("Min. latency / latency: %d/%d" % (log_min_latency[-1], log_latency[-1]))
            print("Total FIFO Size (kB): %d" % log_total_fifo_size[-1])

            iteration += 1

        end_time = time.time()
        duration = int(end_time - start_time)
        print("Done (%d seconds)" % duration)

        return (
            fifo_depths,
            log_total_fifo_size,
            log_interval,
            log_min_latency,
            log_latency,
            duration,
        )

    def determine_start_depth(
        self,
    ):
        # Attempt to determine start depth for all FIFOs automatically.
        # If it doesn't find a working setting start depth must be set manually,
        # potentially on per-FIFO basis.
        start_depth = 1
        last_start_depth = 1
        last_interval = 0
        start_depth_found = False

        while not start_depth_found and not self.error:
            print("Testing start depth of %d" % start_depth)
            self.reset_accelerator()

            # Configure FIFOs
            for i in range(0, self.num_fifos):
                self.configure_fifo(i, mode=1, depth=start_depth)

            # Start accelerator and let it run for a long time
            self.start_accelerator()
            time.sleep(1)

            # Examine performance
            (
                overflow_err,
                underflow_err,
                frame,
                checksum,
                min_latency,
                latency,
                interval,
            ) = self.observe_instrumentation()
            if (
                interval > 0
                and interval == last_interval
                and not overflow_err
                and not underflow_err
            ):
                # Accelerator runs with stable interval, reset to previous start depth
                start_depth_found = True
                start_depth = last_start_depth
            else:
                # Start depth is still too small, increase for next try
                last_start_depth = start_depth
                start_depth = start_depth * 2

            last_interval = interval

            if start_depth > 1000000:
                print("Couldn't find a working start depth, please set manually")
                self.error = True

        # Determine runtime per iteration based on performance, so that stable-state is guaranteed
        # Use a simple overestimation for now to be safe
        iteration_runtime = max(0.01, (min_latency * 5) * 10 / 1000 / 1000 / 1000)

        print("Determined start depth for all FIFOs: %d" % start_depth)
        print("Determined iteration runtime based on performance: %f s" % iteration_runtime)
        return (start_depth, iteration_runtime)


def run_idle(*args, **kwargs):
    # Program FPGA without running accelerator. Only used in the context of power measurement
    runtime = kwargs["runtime"]
    frequency = kwargs["frequency"]
    seed = kwargs["seed"]
    bitfile = kwargs["bitfile"]
    settingsfile = kwargs["settingsfile"]
    devID = kwargs["device"]

    device = Device.devices[devID]

    # TODO: deduplicate code, unify drivers
    if settingsfile != "":
        with open(settingsfile, "r") as f:
            settings = json.load(f)
            if "fclk_mhz" in settings:
                frequency = settings["fclk_mhz"]
            fifo_widths = settings["fifo_widths"]

    print("Programming FPGA..")
    sys.setrecursionlimit(10000)
    PL.reset()
    accel = FINNLiveFIFOOverlay(
        bitfile_name=bitfile, device=device, fclk_mhz=frequency, seed=seed, fifo_widths=fifo_widths
    )
    if accel.error:
        print("Error: Accelerator initialization failed.")
        sys.exit(1)

    print("Running idle for %d seconds.." % runtime)
    time.sleep(runtime)
    print("Done.")


def main(*args, **kwargs):
    frequency = kwargs["frequency"]
    seed = kwargs["seed"]
    bitfile = kwargs["bitfile"]
    reportfile = kwargs["reportfile"]
    settingsfile = kwargs["settingsfile"]
    devID = kwargs["device"]

    device = Device.devices[devID]
    report_dir = os.path.dirname(reportfile)
    folding_config_lfs = None

    # overwrite frequency if specified in settings file
    if settingsfile != "":
        with open(settingsfile, "r") as f:
            settings = json.load(f)
            if "fclk_mhz" in settings:
                frequency = settings["fclk_mhz"]

            # For live FIFO-sizing, we also expect the FIFO widths (in bits) exported by FINN, e.g.,
            # {'fifo_widths': {"0": 8, "1": 32, "2": 24}}
            fifo_widths = settings["fifo_widths"]

            # The settings can also contain the original folding config,
            # into which we can insert the live FIFO sizes once we are done
            if "folding_config_before_lfs" in settings:
                folding_config_lfs = settings["folding_config_before_lfs"]

    print("Programming FPGA..")
    # Increase recursion limit because the default value (1000) caused pickle RecursionErrors
    # during PYNQ cache handling for accelerators with many FIFOs (exact reason unknown)
    sys.setrecursionlimit(10000)
    # Reset PYNQ cache, without this we encountered issues where PYNQ would try to load
    # an incorrect combination of .bit and .hwh file, see https://github.com/Xilinx/PYNQ/issues/1409
    PL.reset()
    accel = FINNLiveFIFOOverlay(
        bitfile_name=bitfile, device=device, fclk_mhz=frequency, seed=seed, fifo_widths=fifo_widths
    )
    if accel.error:
        print("Error: Accelerator initialization failed.")
        sys.exit(1)

    print("Determining start depth..")
    (start_depth, iteration_runtime) = accel.determine_start_depth()

    # First pass
    print("Starting first pass..")
    pass1_result = accel.size_iteratively(start_depth, iteration_runtime)
    (
        fifo_depths,
        log_total_fifo_size,
        log_interval,
        log_min_latency,
        log_latency,
        duration,
    ) = pass1_result

    # Visualize results and save as "fifo_sizing_graph.png"
    fig, ax1 = plt.subplots()

    color = "tab:red"
    ax1.set_xlabel("Iteration")
    ax1.set_ylabel("Total FIFO Size [kB]", color=color)
    ax1.plot(range(len(log_total_fifo_size)), log_total_fifo_size, color=color)
    ax1.tick_params(axis="y", labelcolor=color)
    ax1.set_ylim(0, max(log_total_fifo_size))

    ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis

    color = "tab:blue"
    ax2.set_ylabel("Latency [cycles]", color=color)
    ax2.plot(range(len(log_total_fifo_size)), log_latency, color=color)
    ax2.tick_params(axis="y", labelcolor=color)
    # ax2.set_ylim(0, max(log_latency))

    ax2.axhline(log_min_latency[0], color="green", label="Minimum (1st frame) Latency")
    ax2.legend()

    plt.tight_layout()
    plt.savefig(os.path.join(report_dir, "fifo_sizing_graph.png"), dpi=300)

    # Second pass for fine-tuning
    print("Starting second pass..")
    pass2_result = accel.size_iteratively(fifo_depths, iteration_runtime, reduction_factor=0.95)
    (
        fifo_depths,
        log_total_fifo_size,
        log_interval,
        log_min_latency,
        log_latency,
        duration,
    ) = pass2_result

    # Generate fifo_sizing_report.json
    fifo_report = {
        "error": accel.error,
        "fifo_size_total_kB": log_total_fifo_size[-1],
        "fifo_depths": {},
        "fifo_sizes": {},
        "pass_1": {
            "duration": pass1_result[5],
            "log_total_fifo_size": pass1_result[1],
            "log_interval": pass1_result[2],
            "log_min_latency": pass1_result[3],
            "log_latency": pass1_result[4],
        },
        "pass_2": {
            "duration": pass2_result[5],
            "log_total_fifo_size": pass2_result[1],
            "log_interval": pass2_result[2],
            "log_min_latency": pass2_result[3],
            "log_latency": pass2_result[4],
        },
    }
    for fifo, depth in enumerate(fifo_depths):
        size = (depth + accel.fifo_depth_offset) * accel.fifo_widths[str(fifo)]
        fifo_report["fifo_depths"][fifo] = depth + accel.fifo_depth_offset
        fifo_report["fifo_sizes"][fifo] = size
    with open(os.path.join(report_dir, "fifo_sizing_report.json"), "w") as f:
        json.dump(fifo_report, f, indent=2)

    # Generate fifo_depth_export.json to export FIFO depths for use in FINN
    fifo_depth_export = {}
    for fifo, depth in enumerate(fifo_depths):
        fifo_name = "StreamingFIFO_rtl_%d" % fifo
        fifo_depth_export[fifo_name] = {}
        fifo_depth_export[fifo_name]["depth"] = depth + accel.fifo_depth_offset
    with open(os.path.join(report_dir, "fifo_depth_export.json"), "w") as f:
        json.dump(fifo_depth_export, f, indent=2)

    # Also export directly into original folding config for convenience
    if folding_config_lfs:
        for key in list(folding_config_lfs.keys()):
            if key.startswith("StreamingFIFO"):
                fifo_name = "StreamingFIFO_rtl_%d" % int(key.removeprefix("StreamingFIFO_"))
                # Rename FIFO from StreamingFIFO_* to StreamingFIFO_rtl_*
                folding_config_lfs[fifo_name] = folding_config_lfs.pop(key)
                folding_config_lfs[fifo_name]["depth"] = fifo_depth_export[fifo_name]["depth"]
                folding_config_lfs[fifo_name]["impl_style"] = "rtl"
        with open(os.path.join(report_dir, "folding_config_lfs.json"), "w") as f:
            json.dump(folding_config_lfs, f, indent=2)

    # Generate the usual instrumentation performance report based on final state
    min_latency = log_min_latency[-1]
    latency = log_latency[-1]
    interval = log_interval[-1]
    report = {
        "error": accel.error,
        "checksum": 0,
        "min_latency_cycles": min_latency,
        "latency_cycles": latency,
        "interval_cycles": interval,
        "frequency_mhz": round(accel.fclk_mhz_actual),
        "min_latency_ms": round(min_latency * (1 / (accel.fclk_mhz_actual * 1e6)) * 1e3, 6),
        "latency_ms": round(latency * (1 / (accel.fclk_mhz_actual * 1e6)) * 1e3, 6),
        "throughput_fps": round(1 / (interval * (1 / (accel.fclk_mhz_actual * 1e6)))),
        "min_pipeline_depth": round(min_latency / interval, 2),
        "pipeline_depth": round(latency / interval, 2),
    }
    with open(reportfile, "w") as f:
        json.dump(report, f, indent=2)

    print("Done.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Perform iterative FIFO-Sizing on live FINN accelerator"
    )
    parser.add_argument(
        "--frequency", help="FPGA clock frequency in MHz", type=float, default=100.0
    )
    parser.add_argument("--seed", help="LFSR seed for input data generation", type=int, default=1)
    parser.add_argument("--device", help="FPGA device to be used", type=int, default=0)
    parser.add_argument("--bitfile", help="Name of bitfile", default="finn-accel.bit")
    parser.add_argument(
        "--reportfile",
        help="Name of output .json report file",
        type=str,
        default="measured_performance.json",
    )
    parser.add_argument(
        "--settingsfile", help="Name of optional input .json settings file", type=str, default=""
    )
    args = parser.parse_args()
    main([], vars(args))



"""HLS backend implementation of the Squeeze operator."""
# noqa: Duplicate: The HLS implementation is identical to the Unsqueeze
#  operator, maybe these should be unified...
# fmt: off
# Disable formatter. This is deliberately formatted to stay within 80 characters
# per line. Black, however, formats some lines going beyond this.

# Numpy math and arrays
import numpy as np

# Utility for registering HLSBackend HWCustomOp implementations into the module
# scope
from finn.custom_op.fpgadataflow.hls import register_custom_op

# Base class for specializing HW operators as implemented via HLS
from finn.custom_op.fpgadataflow.hlsbackend import HLSBackend

# The generic HW custom operator version of the operator as a base class
from finn.custom_op.fpgadataflow.squeeze import Squeeze


# HLS Backend specialization of the squeeze operator
@register_custom_op
class Squeeze_hls(Squeeze, HLSBackend):  # noqa: N801
    """HLS backend implementation of the Squeeze operator.

    Removes single-dimension entries from the shape of a tensor using HLS synthesis.
    """

    # CapWords convention
    # Node attributes matching the HLS operator
    def get_nodeattr_types(self):
        """Return the dictionary of node attributes for the HLS Squeeze operator."""
        # Start from parent operator class attributes
        attrs = Squeeze.get_nodeattr_types(self)
        # Add the HLSBackend default attributes on top
        attrs.update(HLSBackend.get_nodeattr_types(self))
        # Add/Specialize implementation specific attributes here...
        # Return the updated attributes dictionary
        return attrs

    # Generates list of C++ includes to be placed at the top of the generated code
    def global_includes(self) -> None:
        """Generate list of C++ includes for the top of the generated code."""
        # Currently nothing to include
        self.code_gen_dict["$GLOBALS$"] = []

    # Generates C++ code of type alias, global constant and macro definitions
    def defines(self, var) -> None:
        """Generate C++ code for type alias, global constant, and macro definitions."""
        # Currently nothing to define
        self.code_gen_dict["$DEFINES$"] = []

    def execute_node(self, context, graph) -> None:
        """Execute node via generic HLSBackend implementation (cppsim/rtlsim)."""
        HLSBackend.execute_node(self, context, graph)

    # Generates C++ code for calling the computation part of the operator
    def docompute(self) -> None:
        """Generate C++ code for the computation part of the operator."""
        # Number of iterations required to process the whole folded input stream
        #   Note: This is all but the PE (last) dimension
        num_iter = np.prod(self.get_folded_output_shape()[:-1])
        # Write the body of the top-level function
        self.code_gen_dict["$DOCOMPUTE$"] = [
            # Repeat for the number of inputs
            f"for(std::size_t i = 0; i < {num_iter}; ++i) {{",
            # Pipeline the steps of this loop
            "#pragma HLS pipeline II=1 style=flp",
            # Just read from the input and immediately write the same element to
            # the output. Squeezed dimensions, i.e., those with a size of 1 do
            # not contribute to the number and order of elements and thus can
            # simply be ignored.
            "out0_V.write(in0_V.read());",
            "}"  # noqa: f-string symmetry
        ]

    # Generates essentially the head of the C++ function from which the IP block
    # will be generated during ipgen, i.e. actual synthesis
    def blackboxfunction(self) -> None:
        """Generate the C++ function signature for the IP block generation."""
        # Insert function head describing the top level interface of the
        # squeeze operator
        self.code_gen_dict["$BLACKBOXFUNCTION$"] = [
            # Note: Assumes stream type aliases to be set in defines
            f"void {self.onnx_node.name} (",
            f"  hls::stream<ap_uint<{self.get_instream_width()}>> &in0_V,",
            f"  hls::stream<ap_uint<{self.get_outstream_width()}>> &out0_V",
            ")",
        ]



"""FINN command-line interface.

This module provides the main CLI for FINN, including commands for building,
running, testing, and managing dependencies and configurations.
"""

from __future__ import annotations

import click
import importlib
import os
import shlex
import subprocess
import sys
from pathlib import Path
from rich.console import Console

from finn.interface import IS_POSIX
from finn.interface.interface_globals import (
    _resolve_settings_path,
    get_settings,
    set_settings,
    settings_found,
    skip_update_by_default,
    write_settings,
)
from finn.interface.interface_utils import (
    assert_path_valid,
    error,
    resolve_build_dir,
    resolve_deps_path,
    resolve_num_workers,
    set_synthesis_tools_paths,
    status,
    warning,
    write_yaml,
)
from finn.interface.manage_deps import install_finnxsi, update_dependencies
from finn.interface.manage_tests import run_test


def _resolve_module_path(name: str, prefix: str = "finn.") -> str:
    """Resolve the path to modules which are not part of the FINN package hierarchy.

    Args:
        name: Module name to resolve
        prefix: Prefix to try if initial import fails

    Returns:
        Absolute path to the module as a string, or empty string if not found
    """
    # Try to import the module via importlib - allows "-" in names and resolve
    # the absolute path to the first candidate location as a string
    try:
        return str(importlib.import_module(name).__path__[0])
    except ModuleNotFoundError:
        # Try a different location if notebooks have not been found, maybe we
        # are in the Git repository root and should look there as well...
        try:
            return str(importlib.import_module(f"{prefix}{name}").__path__[0])
        except ModuleNotFoundError:
            warning(f"Could not resolve {name}. FINN might not work properly.")
    # Return the empty string as a default...
    return ""


def prepare_finn(
    deps: Path | None,
    flow_config: Path,
    build_dir: Path | None,
    num_workers: int,
    is_test_run: bool = False,
    skip_dep_update: bool = False,
) -> None:
    """Prepare a FINN environment.

    This function:
    0. Reads all settings and environment vars
    1. Updates all dependencies
    2. Sets all environment vars

    Args:
        deps: Path to dependencies
        flow_config: File path for context
        build_dir: Build directory path
        num_workers: Number of parallel workers
        is_test_run: Whether this is a test run
        skip_dep_update: Skip dependency update
    """
    # Resolve settings and dependencies, error if this doesnt work
    if not settings_found():
        warning("Settings file could not be found. Using defaults.")
    else:
        sp = _resolve_settings_path()
        status(f"Using settings file at {sp}")
    settings = get_settings(force_update=True)
    deps_path = resolve_deps_path(deps, settings)
    if deps_path is None:
        error("Dependency location could not be resolved!")
        sys.exit(1)
    else:
        status(f"Using dependency path: {deps_path}")
    os.environ["FINN_DEPS"] = str(deps_path.absolute())

    if "PYTHONPATH" not in os.environ.keys():
        os.environ["PYTHONPATH"] = ""

    # Update / Install all dependencies
    if not skip_dep_update:
        update_dependencies(deps_path)
    else:
        warning("Skipping dependency updates!")

    # Check synthesis tools
    set_synthesis_tools_paths()

    # Resolve the build directory
    resolved_build_dir = resolve_build_dir(
        flow_config, build_dir, settings, is_test_run=is_test_run
    )
    if resolved_build_dir is None:
        error("Could not resolve the build directory!")
        sys.exit(1)
    os.environ["FINN_BUILD_DIR"] = str(resolved_build_dir.absolute())
    if not resolved_build_dir.exists():
        resolved_build_dir.mkdir(parents=True)
    status(f"Build directory set to: {resolved_build_dir}")

    # Resolve the number of workers
    workers = resolve_num_workers(num_workers, settings)
    status(f"Using {workers} workers.")
    os.environ["NUM_DEFAULT_WORKERS"] = str(workers)

    # Resolve paths to some not properly packaged components...
    os.environ["FINN_XSI"] = _resolve_module_path("finn_xsi")
    os.environ["FINN_RTLLIB"] = _resolve_module_path("finn-rtllib")
    os.environ["FINN_CUSTOM_HLS"] = _resolve_module_path("custom_hls")
    os.environ["FINN_NOTEBOOKS"] = _resolve_module_path("notebooks", prefix="")
    os.environ["FINN_TESTS"] = _resolve_module_path("tests", prefix="")

    # Install FINN XSI
    finnxsi_status = install_finnxsi()
    if finnxsi_status:
        status("FINN XSI installed successfully.")
    else:
        error("FINN XSI installation failed.")
        sys.exit(1)


@click.group()
def main_group() -> None:
    """FINN command-line interface main group."""


@click.command(help="Build a hardware design")
@click.option("--dependency-path", "-d", default="")
@click.option("--build-path", "-b", help="Specify a build temp path of your choice", default="")
@click.option(
    "--num-workers",
    "-n",
    help="Number of parallel workers for FINN to use. When -1, automatically use 75% of cores",
    default=-1,
    show_default=True,
)
@click.option(
    "--skip-dep-update",
    "-s",
    is_flag=True,
    help="Whether to skip the dependency update. Can be changed in settings via"
    "AUTOMATIC_DEPENDENCY_UPDATES: false",
)
@click.option(
    "--start",
    default="",
    help="If no start_step is given in the dataflow build config, "
    "this starts the flow from the given step.",
)
@click.option(
    "--stop",
    default="",
    help="If no stop_step is given in the dataflow build config, "
    "this stops the flow at the given step.",
)
@click.argument("config")
@click.argument("model")
def build(
    dependency_path: str,
    build_path: str,
    num_workers: int,
    skip_dep_update: bool,
    start: str,
    stop: str,
    config: str,
    model: str,
) -> None:
    """Build a hardware design from a config and model.

    Args:
        dependency_path: Path to FINN dependencies
        build_path: Path for build temporary files
        num_workers: Number of parallel workers (-1 for auto)
        skip_dep_update: Skip dependency update
        start: Starting step for the build flow
        stop: Stopping step for the build flow
        config: Path to build configuration file
        model: Path to model file
    """
    config_path = Path(config).expanduser()
    model_path = Path(model).expanduser()
    build_dir = Path(build_path).expanduser() if build_path != "" else None
    assert_path_valid(config_path)
    assert_path_valid(model_path)
    dep_path = Path(dependency_path).expanduser() if dependency_path != "" else None
    status(f"Starting FINN build with config {config_path.name} and model {model_path.name}!")
    prepare_finn(
        dep_path,
        config_path,
        build_dir,
        num_workers,
        skip_dep_update=(skip_dep_update or skip_update_by_default()),
    )

    # Can import from finn now, since all deps are installed
    # and all environment variables are set correctly
    from finn.builder.build_dataflow import build_dataflow_cfg
    from finn.builder.build_dataflow_config import DataflowBuildConfig

    status("Creating dataflow build config...")
    dfbc: DataflowBuildConfig | None = None
    match config_path.suffix:
        case ".yaml" | ".yml":
            with config_path.open() as f:
                dfbc = DataflowBuildConfig.from_yaml(f.read())
        case ".json":
            with config_path.open() as f:
                dfbc = DataflowBuildConfig.from_json(f.read())
        case _:
            error(
                f"Unknown config file type: {config_path.name}. "
                "Valid formats are: .json, .yml, .yaml"
            )
            sys.exit(1)
    if dfbc is None:
        error("Failed to generate dataflow build config!")
        sys.exit(1)

    # Set start and stop steps
    if dfbc.start_step is None and start != "":
        dfbc.start_step = start
    if dfbc.stop_step is None and stop != "":
        dfbc.stop_step = stop

    # Set output directory to where the config lies, not where FINN lies
    if not Path(dfbc.output_dir).is_absolute():
        dfbc.output_dir = str((config_path.parent / dfbc.output_dir).absolute())
    status(f"Output directory is {dfbc.output_dir}")

    # Add path of config to sys.path so that custom steps can be found
    sys.path.append(str(config_path.parent.absolute()))

    Console().rule(
        f"[bold cyan]Running FINN with config[/bold cyan][bold orange1] "
        f"{config_path.name}[/bold orange1][bold cyan] on model [/bold cyan]"
        f"[bold orange1]{model_path.name}[/bold orange1]"
    )
    build_dataflow_cfg(str(model_path), dfbc)


@click.command(help="Run a script in a FINN environment")
@click.option("--dependency-path", "-d", default="")
@click.option("--build-path", "-b", help="Specify a build temp path of your choice", default="")
@click.option(
    "--skip-dep-update",
    "-s",
    is_flag=True,
    help="Whether to skip the dependency update. Can be changed in settings via"
    "AUTOMATIC_DEPENDENCY_UPDATES: false",
)
@click.option(
    "--num-workers",
    "-n",
    help="Number of parallel workers for FINN to use. When -1, automatically use 75% of cores",
    default=-1,
    show_default=True,
)
@click.argument("script")
def run(
    dependency_path: str, build_path: str, skip_dep_update: bool, num_workers: int, script: str
) -> None:
    """Run a script in a FINN environment.

    Args:
        dependency_path: Path to FINN dependencies
        build_path: Path for build temporary files
        skip_dep_update: Skip dependency update
        num_workers: Number of parallel workers (-1 for auto)
        script: Path to script file to execute
    """
    script_path = Path(script).expanduser()
    build_dir = Path(build_path).expanduser() if build_path != "" else None
    assert_path_valid(script_path)
    dep_path = Path(dependency_path).expanduser() if dependency_path != "" else None
    prepare_finn(
        dep_path,
        script_path,
        build_dir,
        num_workers,
        skip_dep_update=(skip_dep_update or skip_update_by_default()),
    )
    Console().rule(
        f"[bold cyan]Starting script [/bold cyan][bold orange1]{script_path.name}[/bold orange1]"
    )
    subprocess.run(
        shlex.split(f"{sys.executable} {script_path.name}", posix=IS_POSIX), cwd=script_path.parent
    )


@click.command(help="Run a given benchmark configuration.")
@click.option("--bench_config", help="Name or path of experiment configuration file", required=True)
@click.option("--dependency-path", "-d", default="")
@click.option("--num-workers", "-n", default=-1, show_default=True)
@click.option(
    "--build-path",
    "-b",
    help="Specify a build temp path of your choice",
    default="",
)
def bench(bench_config: str, dependency_path: str, num_workers: int, build_path: str) -> None:
    """Run a given benchmark configuration.

    Args:
        bench_config: Name or path of experiment configuration file
        dependency_path: Path to FINN dependencies
        num_workers: Number of parallel workers
        build_path: Path for build temporary files
    """
    console = Console()
    build_dir = Path(build_path).expanduser() if build_path != "" else None
    dep_path = Path(dependency_path).expanduser() if dependency_path != "" else None
    prepare_finn(dep_path, Path(), build_dir, num_workers)
    console.rule("RUNNING BENCHMARK")

    # Late import because we need prepare_finn to setup remaining dependencies first
    from finn.benchmarking.bench import start_bench_run

    exit_code = start_bench_run(bench_config)
    sys.exit(exit_code)


@click.command(help="Run a given test. Uses /tmp/FINN_TMP as the temporary file location")
@click.option(
    "--variant",
    "-v",
    help="Which test to execute (quick, quicktest_ci, full_ci)",
    default="quick",
    show_default=True,
)
@click.option("--dependency-path", "-d", default="")
@click.option("--num-workers", "-n", default=-1, show_default=True)
@click.option("--num-test-workers", "-t", default="auto", show_default=True)
@click.option(
    "--build-path",
    "-b",
    help="Specify a build temp path of your choice",
    default="",
)
def test(
    variant: str, dependency_path: str, num_workers: int, num_test_workers: str, build_path: str
) -> None:
    """Run a given test variant.

    Args:
        variant: Which test to execute (quick, quicktest_ci, full_ci)
        dependency_path: Path to FINN dependencies
        num_workers: Number of parallel workers
        num_test_workers: Number of test workers
        build_path: Path for build temporary files
    """
    console = Console()
    build_dir = Path(build_path).expanduser() if build_path != "" else None
    dep_path = Path(dependency_path).expanduser() if dependency_path != "" else None
    prepare_finn(dep_path, Path(), build_dir, num_workers, is_test_run=True)
    status(f"Using {num_test_workers} test workers")
    console.rule("RUNNING TESTS")
    run_test(variant, num_test_workers)


@click.group(help="Dependency management")
def deps() -> None:
    """Dependency management command group."""


@click.command(help="Update or install dependencies to the given path")
@click.option(
    "--path",
    "-p",
    help="Path to install to",
    default="",
    show_default=True,
)
def update(path: str) -> None:
    """Update or install dependencies to the given path.

    Args:
        path: Path to install dependencies to
    """
    dep_path = Path(path).expanduser() if path != "" else None
    prepare_finn(dep_path, Path(), None, 1)


@click.group(help="Manage FINN settings")
def config() -> None:
    """Manage FINN settings command group."""
    # TODO: Config remove?


def _command_get_settings() -> dict[str, str | int | bool]:
    """Get settings from the settings file.

    Returns:
        Dictionary of settings key-value pairs
    """
    sp = _resolve_settings_path()
    if sp is None:
        error("Could not find a settings file. Stopping")
        sys.exit(1)
    status(f"Found settings at {sp}")
    return get_settings(force_update=True)


@click.command("list", help="List the settings files contents")
def config_list() -> None:
    """List the settings file contents."""
    console = Console()
    for k, v in _command_get_settings().items():
        console.print(f"[blue]{k}[/blue]: {v}")


@click.command("get", help="Get a specific key from the settings")
@click.argument("key")
def config_get(key: str) -> None:
    """Get a specific key from the settings.

    Args:
        key: Setting key to retrieve
    """
    settings = _command_get_settings()
    if key not in settings.keys():
        error(f"Key {key} could not be found in the settings file!")
        sys.exit(1)
    Console().print(f"[blue]{key}[/blue]: {settings[key]}")


@click.command("set", help="Set a key to a given value")
@click.argument("key")
@click.argument("value")
def config_set(key: str, value: str) -> None:
    """Set a key to a given value.

    Args:
        key: Setting key to set
        value: Value to set
    """
    if not settings_found():
        error(
            "Settings file not found. To create a template one at the "
            'default location run "finn config create"'
        )
        sys.exit(1)
    else:
        settings = get_settings(force_update=True)
        settings[key] = value
    set_settings(settings)
    write_settings()


@click.command(
    "create",
    help="Create a template settings file. If one exists at the given path, "
    "its overwritten. Please enter a directory, no filename",
)
@click.argument("path", default="~/.finn/")
def config_create(path: str) -> None:
    """Create a template settings file.

    Args:
        path: Directory path where to create the settings file
    """
    p = Path(path).expanduser()
    if p.suffix != "":
        error("Please specify a path to a directory, not a file!")
        sys.exit(1)
    complete_path = p / "settings.yaml"
    settings = get_settings(force_update=False)
    msettings = {key: str(settings[key]) for key in settings.keys()}
    if not write_yaml(msettings, complete_path):
        error(f"Writing to {complete_path} failed!")
        sys.exit(1)
    status(f"File written to {complete_path}")


def main() -> None:
    """Entry point for the FINN command-line interface."""
    config.add_command(config_list)
    config.add_command(config_create)
    config.add_command(config_get)
    config.add_command(config_set)
    deps.add_command(update)
    main_group.add_command(config)
    main_group.add_command(deps)
    main_group.add_command(build)
    main_group.add_command(bench)
    main_group.add_command(test)
    main_group.add_command(run)
    main_group()


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utilities for Vitis HLS synthesis and IP generation."""

from __future__ import annotations

import os
import re
from pathlib import Path
from subprocess import CalledProcessError

from finn.util.basic import launch_process_helper, which
from finn.util.exception import FINNInternalError, FINNUserError


class CallHLS:
    """Call vitis_hls to run HLS build tcl scripts."""

    def __init__(
        self, tcl_script: str | Path, code_gen_dir: str | Path, ipgen_path: str | Path
    ) -> None:
        """Create a new HLS builder.

        Args:
            tcl_script: Path to the Tcl script that will be executed to build.
            code_gen_dir: Directory that contains the Tcl script (and sources).
            ipgen_path: Path to the IP Generation project.
        """
        self.tcl_script = Path(tcl_script)
        self.ipgen_path = Path(ipgen_path)
        self.code_gen_dir = Path(code_gen_dir)
        self.ipgen_script: Path | None = None
        if self.tcl_script not in self.code_gen_dir.iterdir():
            raise FINNInternalError(
                f"code_gen_dir {code_gen_dir} does not"
                f"seem to contain the Tcl script {tcl_script}"
            )

    def build(self) -> None:
        """Create and run a shell script to execute the Tcl script in code_gen_dir."""
        vivado_path = os.environ.get("XILINX_VIVADO")
        if vivado_path is None:
            raise FINNUserError("XILINX_VIVADO was not set but is required.")
        # xsi kernel lib name depends on Vivado version (renamed in 2024.2)
        match = re.search(r"\b(20\d{2})\.(1|2)\b", vivado_path)
        if match is None:
            raise FINNUserError(f"Could not find a version number in XILINX_VIVADO: {vivado_path}")
        year, minor = int(match.group(1)), int(match.group(2))
        if (year, minor) > (2024, 2):
            if which("vitis-run") is None:
                raise FINNUserError("vitis-run not found in PATH")
            vitis_cmd = f"vitis-run --mode hls --tcl {self.tcl_script}\n"
        else:
            if which("vitis_hls") is None:
                raise FINNUserError("vitis_hls was not found in PATH!")
            vitis_cmd = f"vitis_hls {self.tcl_script}\n"
        self.ipgen_script = self.code_gen_dir / "ipgen.sh"
        working_dir = Path.cwd()
        with self.ipgen_script.open("w") as f:
            f.write("#!/bin/bash \n")
            f.write("set -e\n")
            f.write(f"cd {self.code_gen_dir}\n")
            f.write(vitis_cmd)
            f.write(f"cd {working_dir}\n")
        try:
            launch_process_helper(["bash", str(self.ipgen_script)], print_stdout=False)
        except CalledProcessError as e:
            raise FINNUserError(
                f"HLS IP Generation failed. Logs and sources are in {self.code_gen_dir}"
            ) from e



# Copyright (c) 2020-2022, Xilinx
# Copyright (C) 2022-2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.infer_datatypes import InferDataTypes


class RoundAndClipThresholds(Transformation):
    """For MultiThreshold nodes operating on integer inputs, round up
    thresholds values to the nearest integer. Additionally, if the input
    is unsigned, sets negative thresholds to zero. Type-casts thresholds (back)
    to the float32 container type (this is separate from the quantization
    annotation). Runs InferDataTypes() afterward to propagate any changes to the
    quantization data types."""

    def apply(self, model: ModelWrapper):  # noqa
        graph = model.graph
        graph_modified = False
        for index, node in enumerate(graph.node):
            op_type = node.op_type
            if op_type == "MultiThreshold" or op_type.startswith("Thresholding"):
                thresholds = model.get_initializer(node.input[1])
                if thresholds is None:
                    continue
                dtype = model.get_tensor_datatype(node.input[0])
                # This transformation only applies to thresholding operations
                # operating on integer inputs
                if not dtype.is_integer():
                    continue
                # Round thresholds up to nearest integer and clip thresholds
                # outside the input range
                #   Note: This might promote the thresholds to float64 and
                #   introduce extra inaccuracies due to large integers not being
                #   exactly representable in floating-point representation.
                #   See for example: np.ceil(np.float32(16777217)) == 16777216
                new_thresholds = np.clip(np.ceil(thresholds), dtype.min(), dtype.max() + 1)
                # Convert back to the preferred float32 container type
                new_thresholds = new_thresholds.astype(np.float32)
                # Insert the rounded and clipped thresholds back into the model
                model.set_initializer(node.input[1], new_thresholds)
                # The rounded and clipped thresholds now fit into a data type
                # that is one bit bigger than the input datatype
                # Determine new max_value
                max_val = dtype.max() + 1
                if not dtype.signed():
                    tdt = DataType.get_smallest_possible(max_val)
                else:
                    tdt = DataType.get_smallest_possible(-(max_val) - 1)
                model.set_tensor_datatype(node.input[1], tdt)
                # If hw op we need to set the weight data type attribute as well
                if op_type.startswith("Thresholding"):
                    inst = getCustomOp(node)
                    inst.set_nodeattr("weightDataType", tdt.name)
                # ones
                if np.any(new_thresholds != thresholds):
                    # Track the graph has been modified to inform the transform
                    # container to exhaustively repeat this transformation until
                    # no changes are possible
                    graph_modified = True
                    # Immediately exit here to propagate the data type changes
                    # before considering the next node
                    break
        model = model.transform(InferDataTypes())
        return model, graph_modified



# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import math
import numpy as np
import onnx.numpy_helper as np_helper
import os
import qonnx.custom_op.general.xnorpopcount as xp
import textwrap
from qonnx.core.datatype import DataType
from qonnx.custom_op.general.multithreshold import multithreshold
from qonnx.util.basic import (
    calculate_matvec_accumulator_range,
    interleave_matrix_outer_dim_from_partitions,
    roundup_to_integer_multiple,
)

from finn.custom_op.fpgadataflow.hwcustomop import HWCustomOp
from finn.util.data_packing import numpy_to_hls_code, pack_innermost_dim_as_hex_string
from finn.util.logging import log

# ONNX i/o tensor shape assumptions for MatrixVectorActivation:
# input 0 is the input tensor, shape (.., i_size) = (..., MW)
# input 1 is the weight tensor, shape (i_size, o_size) = (MW, MH)
# (optional) input 2 is the thresholds tensor, shape (o_size, n_thres)
# output 0 is the output tensor, shape (.., o_size) = (..., MH)
# the ... here can be any shape (representing groups of vectors)


class MVAU(HWCustomOp):
    """Abstraction layer for HW implementation of MatrixVectorActivation layers."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            "PE": ("i", True, 0),
            "SIMD": ("i", True, 0),
            "MW": ("i", True, 0),
            "MH": ("i", True, 0),
            "resType": ("s", False, "auto", {"auto", "lut", "dsp"}),
            "ActVal": ("i", False, 0),
            # FINN DataTypes for inputs, weights, outputs
            "inputDataType": ("s", True, ""),
            "weightDataType": ("s", True, ""),
            "outputDataType": ("s", True, ""),
            # FINN DataType for accumulator -- auto-computed and updated
            "accDataType": ("s", False, "INT32"),
            # use xnor-popcount for binary weights/inputs, thus treating them
            # as bipolar
            "binaryXnorMode": ("i", False, 0, {0, 1}),
            # no-activation mode (produce accumulators)
            "noActivation": ("i", False, 0, {0, 1}),
            # number of input vectors, examples:
            # [1] is a single vector (like a FC layer with batch=1)
            # [4] is four vectors (like a FC layer with batch=4)
            # [1, 4, 4] is four * four vectors (like a conv layer with batch=1)
            "numInputVectors": ("ints", False, [1]),
            # memory mode for the FC weights
            # internal_embedded -- embedded weights, long compile/synth times
            # internal_decoupled -- default, streaming weights with streamer packaged inside IP
            # external -- streaming weights with external streamer
            "mem_mode": (
                "s",
                False,
                "internal_decoupled",
                {"internal_embedded", "internal_decoupled", "external"},
            ),
            # FPGA resource type for memories in internal_decoupled mode
            # auto -- let Vivado decide
            # block -- use BRAM
            # distributed -- use LUTRAM
            # ultra -- use UltraRAM (URAM), must have runtime_writeable_weights=1
            # see also https://www.xilinx.com/support/answers/38070.html
            "ram_style": (
                "s",
                False,
                "auto",
                {"auto", "block", "distributed", "ultra"},
            ),
            # FPGA resource type for threshold memories (if noActivation is False)
            # auto -- let Vivado decide
            # block -- use BRAM
            # distributed -- use LUTRAM
            "ram_style_thresholds": (
                "s",
                False,
                "auto",
                {"auto", "block", "distributed"},
            ),
            # (mem_mode = internal_decoupled only) whether weights will be
            # writeable through an AXI-lite interface during runtime
            # 1 for enabled, 0 for disabled.
            # see finn-rtllib/memstream/doc/README for more about the memory
            # address map used for writable weights
            # IMPORTANT: After using AXI lite to either read or write the weights,
            # always "flush" the accelerator by first passing a dummy input
            # vector through the accelerator. This will get rid of any old
            # weight data from the weight FIFOs.
            "runtime_writeable_weights": ("i", False, 0, {0, 1}),
            "pumpedMemory": ("i", False, 0, {0, 1}),
            # dynamic input
            "dynamic_input": ("i", False, 0, {0, 1}),
        }
        my_attrs.update(super().get_nodeattr_types())
        return my_attrs

    def execute_node(self, context, graph):
        node = self.onnx_node
        in_act = context[node.input[0]]
        # ensure that shape is compatible
        in_act = in_act.reshape(self.get_normal_input_shape())

        if self.get_nodeattr("dynamic_input"):
            mvau_w = context[node.input[1]]
        else:
            mvau_w_init = [x for x in graph.initializer if x.name == node.input[1]][0]
            mvau_w = np_helper.to_array(mvau_w_init)

        # Matrix multiplication
        if self.get_nodeattr("binaryXnorMode"):
            # Note: activation/weights are expected to be binary
            # (by design coming from the transformation inferring this operation mode)
            result = xp.xnorpopcountmatmul(in_act, mvau_w)
        elif (
            self.get_nodeattr("inputDataType") == "BIPOLAR"
            and self.get_nodeattr("weightDataType") == "BIPOLAR"
        ):
            # Convert to binary and use xnorpopcountmatmul function
            result = xp.xnorpopcountmatmul((in_act + 1) / 2, (mvau_w + 1) / 2)
        else:
            # Regular matrix multiplication
            result = np.matmul(in_act, mvau_w)
        if self.get_nodeattr("noActivation") == 0:
            mvau_thr_init = [x for x in graph.initializer if x.name == node.input[2]][0]
            mvau_thr = np_helper.to_array(mvau_thr_init)
            odt_is_bipolar = self.get_nodeattr("outputDataType") == "BIPOLAR"
            out_scale = 2 if odt_is_bipolar else 1
            out_bias = -1 if odt_is_bipolar else self.get_nodeattr("ActVal")
            if result.ndim == 4:
                # NHWC to NCHW for multithreshold node
                result = result.transpose((0, 3, 1, 2))
            result = multithreshold(result, mvau_thr, out_scale, out_bias)
            if result.ndim == 4:
                # NCHW to NHWC
                result = result.transpose((0, 2, 3, 1))
        oshape = context[node.output[0]].shape
        context[node.output[0]] = result.reshape(oshape)

    def verify_node(self):
        info_messages = []
        # verify that "backend" is set to "fpgadataflow"
        backend_value = self.get_nodeattr("backend")
        if backend_value == "fpgadataflow":
            info_messages.append("Attribute backend is set correctly")
        else:
            info_messages.append('Attribute backend should be set to "fpgadataflow"')

        # verify that all necessary attributes exist
        # TODO collect automatically from get_nodeattr_types
        try:
            self.get_nodeattr("code_gen_dir_cppsim")
            self.get_nodeattr("executable_path")
            self.get_nodeattr("resType")
            self.get_nodeattr("MW")
            self.get_nodeattr("MH")
            self.get_nodeattr("SIMD")
            self.get_nodeattr("PE")
            self.get_nodeattr("inputDataType")
            self.get_nodeattr("weightDataType")
            self.get_nodeattr("outputDataType")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append("""The required MatrixVectorActivation attributes do not exist.""")

        # verify the number of inputs depending on noActivation value
        # check noActivation value to determine the number of inputs
        no_act = self.get_nodeattr("noActivation")

        if no_act == 1:
            if len(self.onnx_node.input) == 2:
                info_messages.append("The number of inputs is correct")
            else:
                info_messages.append(
                    """MatrixVectorActivation needs in no
                            activation mode 2 inputs (data input and weights)"""
                )
        elif no_act == 0:
            if len(self.onnx_node.input) == 3:
                info_messages.append("The number of inputs is correct")
            else:
                info_messages.append(
                    """MatrixVectorActivation needs 3 inputs
                            (data input and weights and threshold values)"""
                )
        else:
            info_messages.append(
                """noActivation attribute contains {} should
                be 0 or 1""".format(
                    no_act
                )
            )
        return info_messages

    def infer_node_datatype(self, model):
        node = self.onnx_node
        idt = model.get_tensor_datatype(node.input[0])
        if idt != self.get_input_datatype(0):
            warn_str = "inputDataType changing for %s: %s -> %s " % (
                node.name,
                str(self.get_input_datatype(0)),
                str(idt),
            )
            log.warning(warn_str)
        self.set_nodeattr("inputDataType", idt.name)
        # set output datatype from property
        odt = self.get_output_datatype()
        model.set_tensor_datatype(node.output[0], odt)

    def get_input_datatype(self, ind=0):
        """Returns FINN DataType of input."""
        # when performing FIFO insertion on an FC layer with ext weights, the ind
        # parameter can be > 0 (referring to the weights) so handle that here
        if ind == 0:
            return DataType[self.get_nodeattr("inputDataType")]
        elif ind == 1:
            return DataType[self.get_nodeattr("weightDataType")]
        else:
            raise Exception("Undefined input ind for this layer type")

    def get_accumulator_datatype(self):
        """Returns FINN DataType of accumulator"""
        return DataType[self.get_nodeattr("accDataType")]

    def get_output_datatype(self, ind=0):
        """Returns FINN DataType of output."""
        return DataType[self.get_nodeattr("outputDataType")]

    def get_instream_width(self, ind=0):
        if ind == 0:
            i_bits = self.get_input_datatype(0).bitwidth()
            width = i_bits * self.get_nodeattr("SIMD")
        elif ind == 1:
            if self.get_nodeattr("dynamic_input"):
                width = (
                    self.get_folded_input_shape(ind)[-1] * self.get_input_datatype(ind).bitwidth()
                )
            elif (
                self.get_nodeattr("mem_mode") == "internal_decoupled"
                or self.get_nodeattr("mem_mode") == "external"
            ):
                pe = self.get_nodeattr("PE")
                simd = self.get_nodeattr("SIMD")
                wp = self.get_input_datatype(1).bitwidth()
                width = pe * simd * wp
            else:
                width = 0
        elif ind == 2:
            # check if integrated thresholding and return 0
            # because threshold values are always embedded
            # or raise expection if there shouldn't be
            # a third input to the node
            act = not self.get_nodeattr("noActivation")
            if act:
                width = 0
            else:
                raise Exception("Index out of range")
        else:
            raise Exception("Index out of range")
        return width

    def get_outstream_width(self, ind=0):
        o_bits = self.get_output_datatype().bitwidth()
        out_width = o_bits * self.get_nodeattr("PE")
        return out_width

    def get_folded_input_shape(self, ind=0):
        mw = self.get_nodeattr("MW")
        mh = self.get_nodeattr("MH")
        simd = self.get_nodeattr("SIMD")
        pe = self.get_nodeattr("PE")
        sf = mw // simd
        nf = mh // pe
        vecs = list(self.get_nodeattr("numInputVectors"))

        if ind == 0:
            # calculate shape of input 0
            folded_input_shape = tuple(vecs + [sf, simd])
        elif ind == 1:
            if self.get_nodeattr("dynamic_input"):
                # calculate shape of input 1 (weights dynamic)
                folded_input_shape = tuple(vecs[:2] + [mw] + [nf, pe])
            elif self.get_nodeattr("mem_mode") == "external":
                # calculate shape of input 1 (weights static and external)
                folded_input_shape = tuple(vecs + [sf * nf, simd * pe])
            else:
                raise Exception("Undefined input shape for requested input")
        else:
            raise Exception("Undefined input shape for requested input")

        return folded_input_shape

    def get_folded_output_shape(self, ind=0):
        mh = self.get_nodeattr("MH")
        pe = self.get_nodeattr("PE")
        nf = mh // pe
        vecs = list(self.get_nodeattr("numInputVectors"))
        folded_output_shape = tuple(vecs + [nf, pe])
        return folded_output_shape

    def get_normal_input_shape(self, ind=0):
        mw = self.get_nodeattr("MW")
        if ind == 0:
            vecs = list(self.get_nodeattr("numInputVectors"))
            shape = tuple(vecs + [mw])
        elif ind == 1:
            mh = self.get_nodeattr("MH")
            shape = tuple([mw, mh])
        else:
            raise Exception("Undefined input shape for requested input")
        return shape

    def get_normal_output_shape(self, ind=0):
        mh = self.get_nodeattr("MH")
        vecs = list(self.get_nodeattr("numInputVectors"))
        normal_output_shape = tuple(vecs + [mh])
        return normal_output_shape

    def calc_wmem(self):
        """Calculates and returns WMEM."""
        mw = self.get_nodeattr("MW")
        mh = self.get_nodeattr("MH")
        pe = self.get_nodeattr("PE")
        simd = self.get_nodeattr("SIMD")
        assert mh % pe == 0, "Requirement MH divisable by PE is violated."
        assert mw % simd == 0, "Requirement MW divisable by SIMD is violated."
        wmem = mw * mh // (pe * simd)
        return wmem

    def calc_tmem(self):
        """Calculates and returns TMEM."""
        if self.get_nodeattr("noActivation") == 1:
            return 0
        else:
            mh = self.get_nodeattr("MH")
            pe = self.get_nodeattr("PE")
            return mh // pe

    def uram_estimation(self):
        P = self.get_nodeattr("PE")
        Q = self.get_nodeattr("SIMD")
        wdt = self.get_input_datatype(1)
        W = wdt.bitwidth()
        D_in = self.get_nodeattr("MW")
        D_out = self.get_nodeattr("MH")
        omega = (D_in * D_out) / (Q * P)
        mem_width = Q * W * P
        mmode = self.get_nodeattr("mem_mode")
        mstyle = self.get_nodeattr("ram_style")
        if (
            (mmode == "internal_decoupled" and mstyle != "ultra")
            or (mmode == "internal_embedded" and self.calc_wmem() <= 128)
            or (mmode == "external")
        ):
            return 0
        width_multiplier = math.ceil(mem_width / 72)
        depth_multiplier = math.ceil(omega / 4096)
        return width_multiplier * depth_multiplier

    def bram_estimation(self):
        """Calculates resource estimation for BRAM based on:
        - FINN-R: An End-to-End Deep-Learning Framework for Fast
        Exploration of Quantized Neural Networks
        - M. Blott, T. B. Preusser, N. J. Fraser, G. Gambardella, K. O'Brien,
        Y. Umuroglu, M. Leeser and K. Vissers
        - 12. Sep 2018
        """
        # TODO add in/out FIFO contributions
        P = self.get_nodeattr("PE")
        Q = self.get_nodeattr("SIMD")
        wdt = self.get_input_datatype(1)
        W = wdt.bitwidth()
        D_in = self.get_nodeattr("MW")
        D_out = self.get_nodeattr("MH")
        omega = (D_in * D_out) / (Q * P)
        mem_width = Q * W * P
        mmode = self.get_nodeattr("mem_mode")
        mstyle = self.get_nodeattr("ram_style")
        if (
            (mmode == "internal_decoupled" and mstyle in ["distributed", "ultra"])
            or (mmode == "internal_embedded" and self.calc_wmem() <= 128)
            or (mmode == "external")
        ):
            return 0
        # assuming SDP mode RAMB18s (see UG573 Table 1-10)
        # assuming internal_decoupled (RTL) memory,
        # which is more efficient than internal_embedded (HLS)
        if mem_width == 1:
            return math.ceil(omega / 16384)
        elif mem_width == 2:
            return math.ceil(omega / 8192)
        elif mem_width <= 4:
            return (math.ceil(omega / 4096)) * (math.ceil(mem_width / 4))
        elif mem_width <= 9:
            return (math.ceil(omega / 2048)) * (math.ceil(mem_width / 9))
        elif mem_width <= 18 or omega > 512:
            return (math.ceil(omega / 1024)) * (math.ceil(mem_width / 18))
        else:
            return (math.ceil(omega / 512)) * (math.ceil(mem_width / 36))

    def bram_efficiency_estimation(self):
        wdt = self.get_input_datatype(1)
        W = wdt.bitwidth()
        D_in = self.get_nodeattr("MW")
        D_out = self.get_nodeattr("MH")
        bram16_est = self.bram_estimation()
        if bram16_est == 0:
            return 1
        wbits = W * D_in * D_out
        bram16_est_capacity = bram16_est * 36 * 512
        return wbits / bram16_est_capacity

    def uram_efficiency_estimation(self):
        """Function for URAM efficiency estimation: actual parameter storage
        needed divided by the allocated URAM storage (from estimation)"""
        wdt = self.get_input_datatype(1)
        W = wdt.bitwidth()
        D_in = self.get_nodeattr("MW")
        D_out = self.get_nodeattr("MH")
        uram_est = self.uram_estimation()
        if uram_est == 0:
            return 1
        wbits = W * D_in * D_out
        uram_est_capacity = uram_est * 72 * 4096
        return wbits / uram_est_capacity

    def get_exp_cycles(self):
        pe = self.get_nodeattr("PE")
        simd = self.get_nodeattr("SIMD")
        num_inp_vec = self.get_nodeattr("numInputVectors")
        mh = self.get_nodeattr("MH")
        mw = self.get_nodeattr("MW")
        # since mmv != 1 is not supported yet, we set mmv for now to 1
        mmv = 1
        exp_cycles = (mh / pe) * (mw / simd) * np.prod(num_inp_vec) / mmv
        return int(exp_cycles)

    def minimize_accumulator_width(self, model):
        """Minimize the accumulator bit width according to the weight values,
        input data types, and size of dot product"""
        weights = model.get_initializer(self.onnx_node.input[1])
        # since in the calculation the values of the weight matrix are used,
        # for the bipolar case they need to be converted to bipolar
        if self.get_nodeattr("binaryXnorMode"):
            weights = 2 * weights - 1

        thresholds = None
        if len(self.onnx_node.input) > 2:
            thresholds = model.get_initializer(self.onnx_node.input[2])

        idt = self.get_input_datatype(0)

        if not self.get_nodeattr("dynamic_input"):
            (acc_min, acc_max) = calculate_matvec_accumulator_range(weights, idt)

        # if runtime-writeable weights or dynamic input, then the values of the weights can
        # change and we need to use the worst-case values from the datatypes
        if self.get_nodeattr("runtime_writeable_weights") or self.get_nodeattr("dynamic_input"):
            mw = self.get_nodeattr("MW")
            mh = self.get_nodeattr("MH")
            wdt = self.get_input_datatype(1)
            lower_worst = wdt.min() * np.ones((mw, mh))
            lower_range = calculate_matvec_accumulator_range(lower_worst, idt)
            upper_worst = wdt.max() * np.ones((mw, mh))
            upper_range = calculate_matvec_accumulator_range(upper_worst, idt)
            acc_min = min(min(lower_range), min(upper_range))
            acc_max = max(max(lower_range), max(upper_range))

        # if the thresholds can be used to determine range, then adjust the range
        # according to the known values of the thresholds
        if thresholds is not None:
            threshold_tensor = self.get_hw_compatible_threshold_tensor(thresholds)
            # set threshold datatype (and accumulator datatype implicitly)
            min_threshold = thresholds.min()
            max_threshold = thresholds.max()
            # clip threshold values
            if max_threshold > acc_max or min_threshold < acc_min:
                log.warning(f"Clipping some thresholds in {self.onnx_node.name}")
                thresholds = np.clip(thresholds, acc_min, acc_max)
                model.set_initializer(self.onnx_node.input[2], thresholds)
                threshold_tensor = self.get_hw_compatible_threshold_tensor(thresholds)
                min_threshold = thresholds.min()
                max_threshold = thresholds.max()
            acc_min = min(min_threshold, acc_min)
            acc_max = max(max_threshold, acc_max)

        # if the acc_range is always greater than 0, then acc_max <= 2^P - 1
        if acc_min >= 0:
            acc_bit_width = np.log2(acc_max + 1)
            acc_bit_width = math.ceil(acc_bit_width)
            adt = DataType[f"UINT{acc_bit_width}"]
        # if the acc_range is signed, then acc_min >= -2^{P-1} and acc_max <=
        # 2^{P - 1} - 1, which means 2^{P - 1} >= max(-acc_min, 1 + acc_max)
        else:
            _acc_max = max(-acc_min, 1 + acc_max)
            acc_bit_width = np.log2(_acc_max) + 1
            acc_bit_width = math.ceil(acc_bit_width)
            adt = DataType[f"INT{acc_bit_width}"]

        # if activation, assert that the thresholds can be expressed with adt
        if thresholds is not None:
            assert np.vectorize(adt.allowed)(
                threshold_tensor
            ).all(), "Thresholds in %s can't be expressed with type %s" % (
                self.onnx_node.name,
                str(adt),
            )

        # if no activation, output and accumulator datatypes are the same
        if self.get_nodeattr("noActivation"):
            # if this is the last node in the graph, then ensure the datatype is
            # divisibly by 8 bits
            if model.find_direct_successors(self.onnx_node) is None:
                bw = roundup_to_integer_multiple(adt.bitwidth(), 8)
                new_adt_name = adt.name.replace(str(adt.bitwidth()), str(bw))
                adt = DataType[new_adt_name]
            # for no-activation nodes, output dt = acc dt
            self.set_nodeattr("outputDataType", adt.name)
        self.set_nodeattr("accDataType", adt.name)
        return DataType[self.get_nodeattr("accDataType")]

    def minimize_weight_bit_width(self, model):
        """Minimize the bit width based on the values of the weights"""
        if not (
            self.get_nodeattr("runtime_writeable_weights") or self.get_nodeattr("dynamic_input")
        ):
            weights = model.get_initializer(self.onnx_node.input[1])
            w_min = weights.min()
            w_max = weights.max()
            if w_min < 0:
                if abs(w_min) > w_max:
                    wdt = DataType.get_smallest_possible(w_min)
                else:
                    wdt = DataType.get_smallest_possible(-w_max - 1)
            else:
                wdt = DataType.get_smallest_possible(w_max)
            self.set_nodeattr("weightDataType", wdt.name)
        return DataType[self.get_nodeattr("weightDataType")]

    def get_hw_compatible_threshold_tensor(self, orig_thres_matrix):
        """Convert the original numpy weight matrix orig_weight_matrix into
        a form suitable for passing to the hlslib call:
        * ensure MH % PE == 0
        * for bipolar weights&inputs, ensure thresholds are positive
        * interleave rows between PEs
        * reshape into (PE, TMEM, n_thres_steps) and return
        """
        mh = self.get_nodeattr("MH")
        pe = self.get_nodeattr("PE")
        tmem = mh // pe
        assert mh % pe == 0, "Requirement MH divisable by PE is violated."
        assert (
            orig_thres_matrix.ndim == 2
        ), """Threshold matrix dimension is
        not as expected (2)."""
        n_thres_steps = orig_thres_matrix.shape[1]
        inp_is_bipolar = self.get_input_datatype(0) == DataType["BIPOLAR"]
        wt_is_bipolar = self.get_input_datatype(1) == DataType["BIPOLAR"]
        # reinterpret inp/wt as bipolar if bin_xnor_mode is iset
        inp_is_binary = self.get_input_datatype(0) == DataType["BINARY"]
        wt_is_binary = self.get_input_datatype(1) == DataType["BINARY"]
        bin_xnor_mode = self.get_nodeattr("binaryXnorMode") == 1
        inp_is_bipolar = inp_is_bipolar or (inp_is_binary and bin_xnor_mode)
        wt_is_bipolar = wt_is_bipolar or (wt_is_binary and bin_xnor_mode)
        if inp_is_bipolar and wt_is_bipolar:
            # ensure all thresholds are nonnegative
            assert (orig_thres_matrix >= 0).all()
            # ensure all thresholds are integer
            assert (orig_thres_matrix.astype(np.int32) == orig_thres_matrix).all()
        ret = orig_thres_matrix
        # ensure channels = mh , duplicating if necessary
        if ret.shape[0] == 1:
            ret = np.tile(ret, (mh, 1))
        assert ret.shape[0] == mh, "Channels of threshold matrix are not as expected (mh)"
        # distribute rows between PEs
        ret = interleave_matrix_outer_dim_from_partitions(ret, pe)
        assert (
            ret.shape[0] == pe
        ), """First dimension after distribution of the
        rows between PEs is not as expected (pe)"""
        assert (
            ret.shape[1] == tmem
        ), """Second dimension after distribution of the
        rows between PEs is not as expected (tmem)"""
        assert (
            ret.shape[2] == n_thres_steps
        ), """Third dimension after distribution of the
        rows between PEs is not as expected (n_thres_steps)"""
        return ret.reshape(1, pe, tmem, n_thres_steps)

    def get_hw_compatible_weight_tensor(self, orig_weight_matrix):
        """Convert the original numpy weight matrix orig_weight_matrix into
        a form suitable for passing to the hlslib call:
        * ensure MH % PE == 0 and MW % SIMD == 0
        * for bipolar {-1,+1} weights, convert to binary {0, 1}
        * interleave rows between PEs
        * reshape into (1, PE, WMEM, SIMD) and return
        """
        mw = self.get_nodeattr("MW")
        mh = self.get_nodeattr("MH")
        pe = self.get_nodeattr("PE")
        simd = self.get_nodeattr("SIMD")
        wmem = self.calc_wmem()
        assert orig_weight_matrix.shape == (
            mw,
            mh,
        ), """Weights matrix doesn't
        have expected shape (mw, mh)"""
        assert mw % simd == 0, "Requirement MH divisable by SIMD is violated."
        assert mh % pe == 0, "Requirement MH divisable by PE is violated."
        # start by transposing the original weight matrix, since ONNX and
        # finn-hlslib use different assumptions
        # ONNX uses (in_features, out_features) and matmul(x, W)
        # finn-hlslib uses (out_features, in_features) and matmul(W, x)
        ret = orig_weight_matrix.T
        if self.get_input_datatype(1) == DataType["BIPOLAR"]:
            # convert bipolar to binary
            ret = (ret + 1) / 2
        # interleave rows between PEs and reshape
        # distribute rows between PEs
        ret = interleave_matrix_outer_dim_from_partitions(ret, pe)
        # create SIMD as innermost dimension and add a dummy outer dim
        ret = ret.reshape(1, pe, wmem, simd)
        # reverse the SIMD dimension
        ret = np.flip(ret, axis=-1)
        return ret

    def make_weight_file(self, weights, weight_file_mode, weight_file_name):
        """Produce a file containing given weights in appropriate format for this
        layer. This file can be used for either synthesis or run-time reconfig
        of weights.

        Arguments:

        * weights : numpy array with weights to be put into the file
        * weight_file_mode : one of {hls_header, decoupled_verilog_dat,
          decoupled_runtime}
        * weight_file_name : filename for the weight file to be generated

        """
        # convert weights into hlslib/rtllib-compatible format
        weight_tensor = self.get_hw_compatible_weight_tensor(weights)
        export_wdt = self.get_input_datatype(1)
        # we have converted bipolar weights to binary for export,
        # so use it as such for weight generation
        if self.get_input_datatype(1) == DataType["BIPOLAR"]:
            export_wdt = DataType["BINARY"]
        if weight_file_mode == "hls_header":
            weight_hls_code = numpy_to_hls_code(weight_tensor, export_wdt, "weights", True, True)
            # write weights into C++ header file as dictated by finn-hlslib
            f_weights = open(weight_file_name, "w")
            if export_wdt.bitwidth() != 1:
                f_weights.write(
                    "const FixedPointWeights<{},{},{},{}> weights = ".format(
                        self.get_nodeattr("SIMD"),
                        export_wdt.get_hls_datatype_str(),
                        self.get_nodeattr("PE"),
                        self.calc_wmem(),
                    )
                )
            else:
                f_weights.write(
                    "const BinaryWeights<{},{},{}> weights = ".format(
                        self.get_nodeattr("SIMD"),
                        self.get_nodeattr("PE"),
                        self.calc_wmem(),
                    )
                )
            f_weights.write(weight_hls_code)
            f_weights.close()
        elif "decoupled" in weight_file_mode:
            # create a weight stream for various flavors of internal_decoupled mode:
            # transpose weight tensor from (1, PE, WMEM, SIMD) to (1, WMEM, PE, SIMD)
            weight_tensor_unflipped = np.transpose(weight_tensor, (0, 2, 1, 3))
            # reverse SIMD flip for saving weights in .npy
            weight_tensor_simd_flipped = np.flip(weight_tensor_unflipped, axis=-1)
            # PE flip for saving weights in .dat
            weight_tensor_pe_flipped = np.flip(weight_tensor_unflipped, axis=-2)
            # reshape weight tensor (simd_flipped and pe_flipped) to desired shape
            pe = self.get_nodeattr("PE")
            simd = self.get_nodeattr("SIMD")
            # simd_flipped
            weight_tensor_simd_flipped = weight_tensor_simd_flipped.reshape(1, -1, pe * simd)
            weight_tensor_simd_flipped = weight_tensor_simd_flipped.copy()
            # flipped
            weight_tensor_pe_flipped = weight_tensor_pe_flipped.reshape(1, -1, pe * simd)
            weight_tensor_pe_flipped = weight_tensor_pe_flipped.copy()
            if weight_file_mode == "decoupled_npy":
                # save weight stream into npy for cppsim
                np.save(weight_file_name, weight_tensor_simd_flipped)
            elif weight_file_mode == "decoupled_verilog_dat":
                # convert weight values into hexstring
                weight_width = self.get_instream_width(1)
                if self.get_nodeattr("dynamic_input"):
                    weight_width = weight_width * simd
                # pad to nearest 4 bits to get hex strings
                weight_width_padded = roundup_to_integer_multiple(weight_width, 4)
                weight_tensor_pe_flipped = pack_innermost_dim_as_hex_string(
                    weight_tensor_pe_flipped, export_wdt, weight_width_padded, prefix=""
                )
                # add zeroes to pad out file to 1024 entries
                weight_stream = weight_tensor_pe_flipped.flatten()
                weight_stream = weight_stream.copy()
                if self.get_nodeattr("pumpedMemory"):
                    # if pe = simd = 1, known bug, ask user to increase parallelism
                    if pe == simd == 1:
                        raise Exception(
                            """Pumped memory with pe=simd=1 is not supported.
                            Please increase parallelism."""
                        )
                    split_w_stream = np.zeros([weight_stream.shape[0] * 2], dtype=object)
                    k = 0
                    for i in range(len(weight_stream)):
                        weight = weight_stream[i]
                        split_w_stream[k] = weight[len(weight) // 2 :]
                        split_w_stream[k + 1] = weight[: len(weight) // 2]
                        k += 2
                    weight_stream = split_w_stream
                with open(weight_file_name, "w") as f:
                    for val in weight_stream:
                        f.write(val + "\n")
            elif weight_file_mode == "decoupled_runtime":
                # memstream axi-lite interface will map each mem line to
                # one or multiple 32-bit words
                weight_width = self.get_instream_width(1)
                if self.get_nodeattr("dynamic_input"):
                    weight_width = weight_width * simd
                words_per_memwidth = 2 ** math.ceil(math.log2(weight_width / 32))
                if words_per_memwidth < 1:
                    words_per_memwidth = 1
                weight_width_padded = words_per_memwidth * 32
                # first, pack and ensure padding to 32 bits
                weight_tensor_pe_flipped = pack_innermost_dim_as_hex_string(
                    weight_tensor_pe_flipped, export_wdt, weight_width_padded, prefix=""
                )
                weight_stream = weight_tensor_pe_flipped.flatten()
                weight_stream = weight_stream.copy()
                with open(weight_file_name, "w") as f:
                    for val in weight_stream:
                        # split into groups of 8 hex digits (= 32 bits)
                        words_32b = textwrap.wrap(val, 8)
                        words_32b.reverse()
                        for word_32b in words_32b:
                            f.write(word_32b + "\n")
            else:
                raise Exception("Unknown weight_file_mode")

        else:
            raise Exception("Unknown weight_file_mode")

    def generate_params(self, model, path):
        mem_mode = self.get_nodeattr("mem_mode")
        code_gen_dir = path
        if not self.get_nodeattr("dynamic_input"):
            # weights, if not external
            weights = model.get_initializer(self.onnx_node.input[1])

            if mem_mode == "internal_embedded":
                # save hlslib-compatible weights in params.h
                weight_filename = "{}/params.h".format(code_gen_dir)
                self.make_weight_file(weights, "hls_header", weight_filename)
            elif mem_mode == "internal_decoupled" or mem_mode == "external":
                weight_filename_sim = "{}/input_1.npy".format(code_gen_dir)
                # save internal_decoupled weights for cppsim
                self.make_weight_file(weights, "decoupled_npy", weight_filename_sim)
                if mem_mode == "internal_decoupled":
                    # also save weights as Verilog .dat file
                    # This file will be ignored when synthesizing UltraScale memory.
                    weight_filename_rtl = "{}/memblock.dat".format(code_gen_dir)
                    self.make_weight_file(weights, "decoupled_verilog_dat", weight_filename_rtl)
            else:
                raise Exception(
                    """Please set mem_mode to "internal_embedded", "internal_decoupled",
                    or "external", currently no other parameter value is supported!"""
                )

        # save thresholds in thresh.h
        if len(self.onnx_node.input) > 2:
            thresholds = model.get_initializer(self.onnx_node.input[2])
            if thresholds is not None:
                threshold_tensor = self.get_hw_compatible_threshold_tensor(thresholds)
                # use UINT32 threshold export for bipolar times bipolar
                inp_is_bipolar = self.get_input_datatype(0) == DataType["BIPOLAR"]
                wt_is_bipolar = self.get_input_datatype(1) == DataType["BIPOLAR"]
                # reinterpret inp/wt as bipolar if bin_xnor_mode is iset
                inp_is_binary = self.get_input_datatype(0) == DataType["BINARY"]
                wt_is_binary = self.get_input_datatype(1) == DataType["BINARY"]
                bin_xnor_mode = self.get_nodeattr("binaryXnorMode") == 1
                inp_is_bipolar = inp_is_bipolar or (inp_is_binary and bin_xnor_mode)
                wt_is_bipolar = wt_is_bipolar or (wt_is_binary and bin_xnor_mode)
                # get computed threshold datatype from attribute
                tdt = DataType[self.get_nodeattr("accDataType")]

                assert np.vectorize(tdt.allowed)(
                    threshold_tensor
                ).all(), "Thresholds in %s can't be expressed with type %s" % (
                    self.onnx_node.name,
                    str(tdt),
                )
                thresholds_hls_code = numpy_to_hls_code(
                    threshold_tensor, tdt, "thresholds", False, True
                )
                # write thresholds into thresh.h
                f_thresh = open("{}/thresh.h".format(code_gen_dir), "w")
                tdt_hls = tdt.get_hls_datatype_str()
                # use binary to export bipolar activations
                export_odt = self.get_output_datatype()
                if self.get_output_datatype() == DataType["BIPOLAR"]:
                    export_odt = DataType["BINARY"]
                odt_hls = export_odt.get_hls_datatype_str()
                f_thresh.write(
                    "static ThresholdsActivation<{},{},{},{},{},{},{}> threshs \
                    = ".format(
                        self.calc_tmem(),
                        self.get_nodeattr("PE"),
                        threshold_tensor.shape[-1],
                        tdt_hls,
                        odt_hls,
                        self.get_nodeattr("ActVal"),
                        "comp::less_equal<%s, %s>" % (tdt_hls, tdt_hls),
                    )
                )
                f_thresh.write(thresholds_hls_code)
                f_thresh.close()

    def get_op_and_param_counts(self):
        in_features = self.get_nodeattr("MW")
        out_features = self.get_nodeattr("MH")
        weight_bits = self.get_input_datatype(1).bitwidth()
        inp_bits = self.get_input_datatype(0).bitwidth()
        num_inp_vec = self.get_nodeattr("numInputVectors")
        num_repetitions = int(np.prod(num_inp_vec))
        mac_count = in_features * out_features * num_repetitions
        # cannonicalize op type: highest bitwidth operand first s.t.
        # e.g. mac_8bx4b and mac_4bx8b don't appear as two different op types
        bw1 = min(inp_bits, weight_bits)
        bw2 = max(inp_bits, weight_bits)
        mac_op_type = "op_mac_%dbx%db" % (bw1, bw2)
        weight_param_type = "param_weight_%db" % (weight_bits)
        weight_count = in_features * out_features
        ret_dict = {mac_op_type: mac_count, weight_param_type: weight_count}
        if self.get_nodeattr("noActivation") == 0:
            tdt = DataType[self.get_nodeattr("accDataType")]
            thres_bits = tdt.bitwidth()
            thres_param_type = "param_threshold_%db" % (thres_bits)
            thres_count = out_features
            ret_dict[thres_param_type] = thres_count
        return ret_dict

    def derive_characteristic_fxns(self, period):
        n_inps = np.prod(self.get_folded_input_shape()[:-1])
        io_dict = {
            "inputs": {
                "in0": [0 for i in range(n_inps)],
            },
            "outputs": {"out0": []},
        }
        mem_mode = self.get_nodeattr("mem_mode")
        if mem_mode in ["internal_decoupled", "external"]:
            n_weight_inps = self.calc_wmem()
            num_w_reps = np.prod(self.get_nodeattr("numInputVectors"))
            io_dict["inputs"]["in1"] = [0 for i in range(num_w_reps * n_weight_inps)]
        super().derive_characteristic_fxns(period, override_rtlsim_dict=io_dict)

    def get_verilog_top_module_intf_names(self):
        intf_names = super().get_verilog_top_module_intf_names()
        try:
            pumped_compute = self.get_nodeattr("pumpedCompute")
        except AttributeError:
            pumped_compute = 0

        if pumped_compute or self.get_nodeattr("pumpedMemory"):
            intf_names["clk2x"] = ["ap_clk2x"]

        dynamic_input = self.get_nodeattr("dynamic_input")
        mem_mode = self.get_nodeattr("mem_mode")
        if dynamic_input:
            weight_width = self.get_instream_width(1)
            weight_width = weight_width * self.get_nodeattr("SIMD")
            intf_names["s_axis"].append(("in1_V", roundup_to_integer_multiple(weight_width, 8)))
        else:
            if mem_mode == "external":
                intf_names["s_axis"].append(("in1_V", self.get_instream_width_padded(1)))
            elif mem_mode == "internal_decoupled":
                # only expose axilite interface if attribute is set
                runtime_writeable = self.get_nodeattr("runtime_writeable_weights")
                if runtime_writeable:
                    intf_names["axilite"] = ["s_axilite"]
        return intf_names

    def code_generation_ipi(self):
        source_target = "./ip/verilog/rtl_ops/%s" % self.onnx_node.name
        cmd = ["file mkdir %s" % source_target]
        dyn_input = self.get_nodeattr("dynamic_input")
        mem_mode = self.get_nodeattr("mem_mode")
        sname = "V"

        # check if additional components are needed
        if dyn_input or mem_mode == "internal_decoupled":
            node_name = self.onnx_node.name
            # create a hierarchy for this layer, with the same port names
            clk_name = self.get_verilog_top_module_intf_names()["clk"][0]
            rst_name = self.get_verilog_top_module_intf_names()["rst"][0]
            dout_name = self.get_verilog_top_module_intf_names()["m_axis"][0][0]
            din_name = self.get_verilog_top_module_intf_names()["s_axis"][0][0]
            cmd.append("create_bd_cell -type hier %s" % node_name)
            # clock and reset
            cmd.append("create_bd_pin -dir I -type clk /%s/%s" % (node_name, clk_name))
            cmd.append("create_bd_pin -dir I -type rst /%s/%s" % (node_name, rst_name))
            # if we need a 2x clock for either compute or memory, instantiate the 2x clk port
            try:
                pumped_compute = self.get_nodeattr("pumpedCompute")
            except AttributeError:
                pumped_compute = 0
            if pumped_compute or self.get_nodeattr("pumpedMemory"):
                clk2x_name = self.get_verilog_top_module_intf_names()["clk2x"][0]
                cmd.append("create_bd_pin -dir I -type clk /%s/%s" % (node_name, clk2x_name))
            else:
                clk2x_name = None
            # streams
            cmd.append(
                "create_bd_intf_pin -mode Master "
                "-vlnv xilinx.com:interface:axis_rtl:1.0 /%s/%s" % (node_name, dout_name)
            )
            cmd.append(
                "create_bd_intf_pin -mode Slave "
                "-vlnv xilinx.com:interface:axis_rtl:1.0 /%s/%s" % (node_name, din_name)
            )
            # instantiate the RTL block
            # Instantiate either the HLS or RTL IP depending on operator
            self.instantiate_ip(cmd)
            # connect MVAU
            cmd.append(
                "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/%s]"
                % (node_name, rst_name, node_name, node_name, rst_name)
            )
            cmd.append(
                "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/%s]"
                % (node_name, clk_name, node_name, node_name, clk_name)
            )
            cmd.append(
                "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                "[get_bd_intf_pins %s/%s/%s]"
                % (node_name, din_name, node_name, node_name, din_name)
            )
            cmd.append(
                "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                "[get_bd_intf_pins %s/%s/%s]"
                % (node_name, dout_name, node_name, node_name, dout_name)
            )

            code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen")
            if dyn_input:
                # dynamic loader
                dynld_rtllib_dir = os.path.join(os.environ["FINN_RTLLIB"], "dynload/hdl/")
                file_suffix = "_dynamic_load_wrapper.v"
                # automatically find memstream verilog component in code generation directory
                for fname in os.listdir(code_gen_dir):
                    if fname.endswith(file_suffix):
                        dynld_tmpl = fname
                dynld_tmpl_name = dynld_tmpl[:-2]
                sourcefiles = [
                    os.path.join(code_gen_dir, dynld_tmpl),
                    dynld_rtllib_dir + "ram_p_c.sv",
                    dynld_rtllib_dir + "dynamic_load.sv",
                ]
                for f in sourcefiles:
                    cmd += ["add_files -copy_to %s -norecurse %s" % (source_target, f)]
                dynld_inst = node_name + "_wdynld"
                # instantiate the cell
                cmd.append(
                    "create_bd_cell -type hier -reference %s /%s/%s"
                    % (dynld_tmpl_name, node_name, dynld_inst)
                )
                # additional dynamic input
                win_name = self.get_verilog_top_module_intf_names()["s_axis"][1][0]
                cmd.append(
                    "create_bd_intf_pin -mode Slave "
                    "-vlnv xilinx.com:interface:axis_rtl:1.0 /%s/%s" % (node_name, win_name)
                )
                # connect
                cmd.append(
                    "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/ap_clk]"
                    % (node_name, clk_name, node_name, dynld_inst)
                )
                cmd.append(
                    "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/ap_rst_n]"
                    % (node_name, rst_name, node_name, dynld_inst)
                )
                cmd.append(
                    "connect_bd_intf_net [get_bd_intf_pins %s/%s/m_axis_0] "
                    "[get_bd_intf_pins %s/%s/in1_%s]"
                    % (node_name, dynld_inst, node_name, node_name, sname)
                )
                cmd.append(
                    "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                    "[get_bd_intf_pins %s/%s/s_axis_0]"
                    % (node_name, win_name, node_name, dynld_inst)
                )
            else:
                # memstream
                runtime_writable = self.get_nodeattr("runtime_writeable_weights") == 1
                axi_dir = os.path.join(os.environ["FINN_RTLLIB"], "axi/hdl/")
                ms_rtllib_dir = os.path.join(os.environ["FINN_RTLLIB"], "memstream/hdl/")
                file_suffix = "_memstream_wrapper.v"
                # automatically find memstream verilog component in code generation directory
                for fname in os.listdir(code_gen_dir):
                    if fname.endswith(file_suffix):
                        strm_tmpl = fname
                strm_tmpl_name = strm_tmpl[:-2]
                sourcefiles = [
                    os.path.join(code_gen_dir, strm_tmpl),
                    axi_dir + "axilite.sv",
                    ms_rtllib_dir + "memstream_axi.sv",
                    ms_rtllib_dir + "memstream.sv",
                ]
                for f in sourcefiles:
                    cmd += ["add_files -copy_to %s -norecurse %s" % (source_target, f)]
                strm_inst = node_name + "_wstrm"
                # instantiate the cell
                cmd.append(
                    "create_bd_cell -type hier -reference %s /%s/%s"
                    % (strm_tmpl_name, node_name, strm_inst)
                )
                # connect
                cmd.append(
                    "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/ap_clk]"
                    % (node_name, clk_name, node_name, strm_inst)
                )
                cmd.append(
                    "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/ap_rst_n]"
                    % (node_name, rst_name, node_name, strm_inst)
                )
                # if using 2x pumped memory, connect the memstreamer's 2x clk input
                # to the 2x clock port. otherwise connect it to the regular clock port.
                if self.get_nodeattr("pumpedMemory"):
                    cmd.append(
                        "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/ap_clk2x]"
                        % (node_name, clk2x_name, node_name, strm_inst)
                    )
                else:
                    cmd.append(
                        "connect_bd_net [get_bd_pins %s/%s] [get_bd_pins %s/%s/ap_clk2x]"
                        % (node_name, clk_name, node_name, strm_inst)
                    )
                cmd.append(
                    "connect_bd_intf_net [get_bd_intf_pins %s/%s/m_axis_0] "
                    "[get_bd_intf_pins %s/%s/in1_%s]"
                    % (node_name, strm_inst, node_name, node_name, sname)
                )
                # runtime writeable weights
                if runtime_writable:
                    axilite_name = self.get_verilog_top_module_intf_names()["axilite"][0]
                    cmd.append(
                        "create_bd_intf_pin -mode Slave "
                        "-vlnv xilinx.com:interface:aximm_rtl:1.0 /%s/%s"
                        % (node_name, axilite_name)
                    )
                    cmd.append(
                        "connect_bd_intf_net [get_bd_intf_pins %s/%s] "
                        "[get_bd_intf_pins %s/%s/%s]"
                        % (node_name, axilite_name, node_name, strm_inst, axilite_name)
                    )
                    # TODO calculate and pass in segment size here
                    cmd.append("assign_bd_address")

            # save bd
            cmd.append("save_bd_design")
        elif mem_mode == "internal_embedded" or mem_mode == "external":
            # base class impl sufficient for internal_embedded/external modes
            self.instantiate_ip(cmd)
        else:
            raise Exception("Unrecognized mem_mode for MatrixVectorActivation")
        return cmd



"""Create C++ and PYNQ drivers for FINN-generated accelerators."""

# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import multiprocessing
import numpy as np
import os
import qonnx
import shlex
import shutil
import subprocess
import sys
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from string import Template
from typing import Dict, List, Optional, Tuple

import finn.util
from finn.builder.build_dataflow_config import FpgaMemoryType
from finn.templates import get_templates_folder
from finn.util.basic import make_build_dir
from finn.util.data_packing import get_driver_shapes, to_external_tensor
from finn.util.exception import FINNInternalError, FINNUserError
from finn.util.logging import log

from . import template_driver


def update_bitfile_path_after_copy(bitfile_path: str, json_path: str) -> None:
    """
    Update the xclbinPath in the JSON configuration to point to the new bitfile location.

    Args:
        json_path (str): Path to the JSON configuration file
        bitfile_path (str): New path to the bitfile (.xclbin)
    """
    if json_path is None or not os.path.exists(json_path):
        raise FINNInternalError("JSON configuration file does not exist or is not specified.")
    if bitfile_path is None or not os.path.exists(bitfile_path):
        raise FINNInternalError("Bitfile path does not exist or is not specified.")
    if not json_path.endswith(".json"):
        raise FINNInternalError("Provided path is not a JSON file.")

    # Read the current JSON configuration
    with open(json_path, "r") as f:
        data = json.load(f)

    # Update the xclbinPath for each device in the configuration
    for device_config in data:
        device_config["xclbinPath"] = os.path.abspath(bitfile_path)

    # Write the updated configuration back to the file
    with open(json_path, "w") as f:
        json.dump(data, f, indent=4)


class MakeCPPDriver(Transformation):
    """Create CPP code to correctly interface the generated
    accelerator, including data packing/unpacking. Should be called
    after conversion to HLS layers, folding and the creation of
    dataflow partitions for correct operation.
    platform: has to be "alveo", otherwise an error is thrown
    Outcome if successful: sets the cpp_driver_dir attribute in the ONNX
    ModelProto's metadata_props field, with the created driver dir as the
    value.
    runtime writeable weights not yet supported.
    """

    # TODO: Enable multiple input types! Now only assumes the first one
    def resolve_dt_name(s: str) -> str:
        """Resolve datatype name for C++ driver code generation.

        Args:
            s: Datatype string to resolve

        Returns:
            Resolved C++ datatype name

        Raises:
            FINNInternalError: If datatype is unknown
        """
        s = s.replace("DataType[", "").replace("]", "")
        if s in ["BINARY", "TERNARY", "BIPOLAR"]:
            return "Datatype" + s[0] + s[1:].lower()
        elif s.startswith("U"):
            return "DatatypeUint<" + s.replace("UINT", "") + ">"
        elif s.startswith("I"):
            return "DatatypeInt<" + s.replace("INT", "") + ">"
        elif "FLOAT" in s:
            return "DatatypeFloat<" + s.replace("FLOAT", "") + ">"
        elif "FIXED" in s:
            return "DatatypeFixed" + s.replace("FIXED", "")
        else:
            raise FINNInternalError(f"Unknown datatype for C++ Driver:{s}")

    def __init__(
        self,
        platform: str,
        version: str,
        host_mem: str,
    ):
        """Initialize MakeCPPDriver transformation.

        Args:
            platform: Target platform (must be "alveo")
            version: Version of finn-cpp-driver to use ("latest" or commit hash)
            host_mem: Memory type (FpgaMemoryType.HOST_MEM or FpgaMemoryType.DEVICE_MEM)

        Raises:
            FINNUserError: If platform is not "alveo"
        """
        super().__init__()
        self.platform: str = platform

        if platform != "alveo":
            raise FINNUserError(
                "CPP driver only supported for Alveo devices, please use PYNQ driver instead."
            )
        self.version = version

        # Define variables for the repository URL and commit hash
        self.repository_url = "https://github.com/eki-project/finn-cpp-driver.git"
        if version == "latest" or version is None:
            self.commit_hash = "HEAD"
        else:
            self.commit_hash = version

        if host_mem == FpgaMemoryType.HOST_MEM:
            self.host_memory = True
        else:
            self.host_memory = False

    def apply(self, model: ModelWrapper) -> Tuple[ModelWrapper, bool]:
        """Apply the MakeCPPDriver transformation to generate C++ driver code.

        Args:
            model: ONNX model wrapper

        Returns:
            Tuple of (modified model, transformation success flag)
        """
        driver_shapes: Dict = get_driver_shapes(model)
        ext_weight_dma_cnt: int  # noqa
        weights_dir: str  # noqa
        # TODO: Enable weight file generation
        # ext_weight_dma_cnt, weights_dir = write_weights(model, cpp_driver_dir)

        # Create a temporary directory for the generated C++ driver code
        cpp_driver_dir = make_build_dir(prefix="cpp_driver_")
        # Store the driver directory path in model metadata
        model.set_metadata_prop("cpp_driver_dir", cpp_driver_dir)
        # Get the path to the FPGA bitstream from model metadata
        xclbin_path = model.get_metadata_prop("bitfile_output")
        # Define paths for configuration files
        json_path = os.path.join(cpp_driver_dir, "acceleratorconfig.json")
        header_path = os.path.join(cpp_driver_dir, "AcceleratorDatatypes.h")

        # Helper function to execute shell commands safely with error handling
        def run_command(command, cwd=None, debug=False):
            """Execute a shell command with error handling.

            Args:
                command: Shell command string to execute.
                cwd: Working directory for command execution. Defaults to None (current directory).
                debug: If True, print command output for debugging. Defaults to False.

            Raises:
                subprocess.CalledProcessError: If the command returns a non-zero exit code.
            """
            try:
                result = subprocess.run(
                    shlex.split(command), cwd=cwd, check=True, text=True, capture_output=True
                )
                if debug:
                    # Print the output for debugging purposes
                    print(result.stdout)
            except subprocess.CalledProcessError as e:
                print(f"Error running command: {command}")
                print(f"Output:{e.stdout}; Error:{e.stderr}")
                raise e

        # Clone and set up the C++ driver repository
        log.info("Downloading C++ driver template...")
        # Initialize git repo and fetch specified version
        run_command("git init", cwd=cpp_driver_dir)
        run_command(f"git remote add origin {self.repository_url}", cwd=cpp_driver_dir)
        run_command(f"git fetch origin {self.commit_hash} --depth=1", cwd=cpp_driver_dir)
        run_command("git checkout FETCH_HEAD", cwd=cpp_driver_dir)
        # Initialize and update all git submodules
        run_command("git submodule update --init --recursive", cwd=cpp_driver_dir)

        log.info("Generating template files...")
        # Check if multiple different input/output types are used.
        if len(set(driver_shapes["idt"])) > 1 or len(set(driver_shapes["odt"])) > 1:
            raise RuntimeError(
                "Multiple different input/output types for the C++ driver\
                    are currently not supported."
            )

        # * Writing the header file
        inputDatatype: str = MakeCPPDriver.resolve_dt_name(
            driver_shapes["idt"][0].replace("'", "")
        )  # .get_canonical_name())
        outputDatatype: str = MakeCPPDriver.resolve_dt_name(
            driver_shapes["odt"][0].replace("'", "")
        )  # .get_canonical_name())
        with open(
            os.path.join(
                cpp_driver_dir, "src", "FINNCppDriver", "config", "FinnDriverUsedDatatypes.h.in"
            ),
            "r",
        ) as f_in:
            header = f_in.read()
            template_handler = Template(header)
            templated_str = template_handler.substitute(
                inputDatatype=inputDatatype, outputDatatype=outputDatatype
            )
            with open(header_path, "w+") as f:
                f.write(templated_str)

        # * Writing the json file
        # TODO: Update this for multi-fpga usage (more than one device!)
        # Path of the xclbin in the finn compiler project
        # Get kernel names using xclbinutil

        if shutil.which("xclbinutil") is None:
            raise RuntimeError(
                "xclbinutil not in PATH or not installed.\
                Required to read kernel names for driver config!"
            )

        # Extract IP layout information from the FPGA bitstream
        # Use xclbinutil to dump the IP layout section from the bitstream to a JSON file
        run_command(
            f"xclbinutil -i {xclbin_path} --dump-section IP_LAYOUT:JSON:ip_layout.json --force",
            cwd=os.path.dirname(xclbin_path),
        )
        # Load the IP layout information from the generated JSON file
        ips = None
        with open(os.path.join(os.path.dirname(xclbin_path), "ip_layout.json")) as f:
            ips = json.loads(f.read())["ip_layout"]["m_ip_data"]

        # Define a filter function to identify input/output DMA kernels
        # Filters for kernels that have valid base addresses
        # and contain "idma" or "odma" in their names
        isIO = (
            lambda x: x["m_type"] == "IP_KERNEL"
            and x["m_base_address"] != "not_used"
            and ("idma" in x["m_name"] or "odma" in x["m_name"])
        )
        # Extract lists of input and output DMA kernel names
        idmas = [x["m_name"] for x in ips if isIO(x) and "idma" in x["m_name"]]
        odmas = [x["m_name"] for x in ips if isIO(x) and "odma" in x["m_name"]]

        # Helper function to format kernel names for the driver configuration
        # Converts "kernelName:instance" to "kernelName:{instance}" format
        def formatKernelName(kname: str):
            """Format kernel name into Vitis-compatible format.

            Args:
                kname: Kernel name string in "name:instance" format.

            Returns:
                Formatted kernel name as "name:{instance}".
            """
            kparts = kname.split(":")
            return kparts[0] + ":{" + kparts[1] + "}"

        # Create JSON configuration entries for input and output DMAs
        jsonIdmas = []
        jsonOdmas = []
        # Map driver's idma names to actual kernels and include shape information
        for i in range(len(driver_shapes["idma_names"])):
            jsonIdmas.append(
                {
                    "kernelName": [
                        formatKernelName(name)
                        for name in idmas
                        if driver_shapes["idma_names"][i] in name
                    ][0],
                    "normalShape": driver_shapes["ishape_normal"][i],
                    "foldedShape": driver_shapes["ishape_folded"][i],
                    "packedShape": driver_shapes["ishape_packed"][i],
                }
            )
        # Map driver's odma names to actual kernels and include shape information
        for i in range(len(driver_shapes["odma_names"])):
            jsonOdmas.append(
                {
                    "kernelName": [
                        formatKernelName(name)
                        for name in odmas
                        if driver_shapes["odma_names"][i] in name
                    ][0],
                    "normalShape": driver_shapes["oshape_normal"][i],
                    "foldedShape": driver_shapes["oshape_folded"][i],
                    "packedShape": driver_shapes["oshape_packed"][i],
                }
            )

        # Create the final JSON configuration structure
        data = []
        data.append(
            {
                # Specify which XRT device to use (0 = first device)
                "xrtDeviceIndex": 0,
                # Store the absolute path to the bitstream
                "xclbinPath": os.path.abspath(xclbin_path),
                "name": "MainDevice",  # Assign a name to this device configuration
                "idmas": jsonIdmas,  # Include the input DMA configurations
                "odmas": jsonOdmas,  # Include the output DMA configurations
            }
        )
        # Write the complete configuration to the JSON file
        with open(json_path, "w+") as f:
            f.write(json.dumps(data, indent=4))

        log.info("Created runtime json config file")

        # Helper function to configure CMake build system
        def configure_cmake(
            source_dir: str,  # Directory containing CMakeLists.txt
            build_dir: str,  # Directory where build files will be generated
            # Additional CMake arguments as string
            cmake_args: Optional[str] = None,
            # Command to invoke CMake
            cmake_executable: str = f"{sys.executable} -m cmake",
        ):
            """Configure CMake build system for the C++ driver.

            Args:
                source_dir: Directory containing the CMakeLists.txt file.
                build_dir: Directory where CMake build files will be generated.
                cmake_args: Additional CMake configuration arguments. Defaults to None.
                cmake_executable: Command to invoke CMake. Defaults to Python's cmake module.
            """
            # Create build directory if it doesn't exist
            os.makedirs(build_dir, exist_ok=True)
            # Split the cmake executable command into arguments
            args = shlex.split(cmake_executable)
            # Add any additional CMake arguments if provided
            if cmake_args:
                cmake_args = shlex.split(cmake_args)
                args.extend(cmake_args)
            # Set CMake policy version to ensure compatibility
            # Needed because CMake 4.0.2 is installed by FINN+ and set minimum version
            # requirements are not correctly picked up by CMake
            args.append("-DCMAKE_POLICY_VERSION_MINIMUM=3.5")
            args.append(os.path.abspath(source_dir))
            log.info(f"Configuring with: {' '.join(args)}")
            result = subprocess.run(args, cwd=build_dir, capture_output=True, text=True)
            if result.returncode != 0:
                log.critical(f"Configure failed with error:\n{result.stderr}")
                raise subprocess.CalledProcessError(
                    result.returncode, args, result.stdout, result.stderr
                )

        # Helper function to build the configured CMake project
        def build_cmake(
            build_dir: str,  # Directory containing the configured build files
            # Build tool to use (default: make)
            cmake_executable: str = "make",
            # Specific target to build (if any)
            build_target: Optional[str] = None,
            # Additional build arguments
            build_args: Optional[List[str]] = None,
        ):
            """Build the configured CMake project.

            Args:
                build_dir: Directory containing the configured build files.
                cmake_executable: Build tool to use. Defaults to "make".
                build_target: Specific target to build. Defaults to None (builds all).
                build_args: Additional build arguments. Defaults to None.

            Raises:
                subprocess.CalledProcessError: If the build fails.
            """
            # Prepare the build command with the executable
            args = [cmake_executable]
            # Add optional build target if specified
            if build_target:
                args += [build_target]
            # Add any additional build arguments
            if build_args:
                args.extend(build_args)
            log.info(f"Building with:{' '.join(args)}")
            # Execute the build command
            result = subprocess.run(args, cwd=build_dir, capture_output=True, text=True)
            # Handle build failures
            if result.returncode != 0:
                log.critical(f"Build failed with error:\n{result.stderr}")
                raise subprocess.CalledProcessError(
                    result.returncode, args, result.stdout, result.stderr
                )

        host_memory_usage = "ON" if self.host_memory else "OFF"

        # Define CMake configuration options for the driver build
        # - Release build type for optimized performance
        # - Disable sanitizers for production builds
        # - Set custom header location
        # - Disable documentation generation
        # - Enable/Disable host memory usage
        cmake_args = f"-DCMAKE_BUILD_TYPE=Release -DFINN_ENABLE_SANITIZERS=Off\
        -DFINN_HEADER_LOCATION={os.path.abspath(header_path)} -DFINN_BUILD_DOC=Off\
            -DFINN_USE_HOST_MEM={host_memory_usage}"

        # Configure the CMake project
        configure_cmake(
            source_dir=cpp_driver_dir,
            build_dir=os.path.join(cpp_driver_dir, "build"),
            cmake_args=cmake_args,
        )
        # Determine optimal number of build threads based on CPU cores
        num_cores = multiprocessing.cpu_count()
        build_cmake(
            build_dir=os.path.join(cpp_driver_dir, "build"), build_args=["-j", str(num_cores)]
        )

        # Helper function to verify that the built driver uses the correct datatypes
        def check_finn_types(bin_dir: str, expectedInputType: str, expectedOutputType: str) -> None:
            """Verify that compiled driver's datatypes match expected types.

            Args:
                bin_dir: Directory containing the finnhpc executable.
                expectedInputType: Expected input datatype string.
                expectedOutputType: Expected output datatype string.

            Raises:
                subprocess.CalledProcessError: If the datatype check command fails.
                RuntimeError: If the actual datatypes don't match expected types.
            """
            # Run the built finnhpc executable with the --check flag to output datatype information
            result = subprocess.run(
                "./finnhpc --check".split(), cwd=bin_dir, capture_output=True, text=True
            )
            if result.returncode != 0:
                log.critical(f"Running datatype check failed with error:\n{result.stderr}")
                raise subprocess.CalledProcessError(result.returncode, result.stdout, result.stderr)
            output = result.stdout
            output_lines = output.splitlines()

            # Verify that the compiled driver's datatypes match the expected types
            # First line contains input type, second line contains output type
            if (
                expectedInputType not in output_lines[0]
                or expectedOutputType not in output_lines[1]
            ):
                log.error(
                    f"FINN types check failed. Expected Types: {expectedInputType},\
                        {expectedOutputType}"
                )
                log.error(f"                           Actual Types: {output}")
                raise FINNInternalError(
                    "Expected C++ driver types to match\
                    expected types."
                )

        # Make the compiled finnhpc executable file executable (chmod +x)
        os.chmod(os.path.join(cpp_driver_dir, "build", "bin", "finnhpc"), 0o755)

        # Verify that the driver was compiled with the correct datatypes
        check_finn_types(
            bin_dir=os.path.join(cpp_driver_dir, "build", "bin"),
            expectedInputType=inputDatatype,
            expectedOutputType=outputDatatype,
        )

        # TODO: Generating weight files
        # weights_dir = output_dir + "/runtime_weights"

        # os.makedirs(weights_dir)
        # idma_idx = 0
        # ext_weight_dma_cnt = 0

        # for node in model.graph.node:
        #     assert (
        #         node.op_type == "StreamingDataflowPartition"
        #     ), "CreateDataflowPartition needs to be applied before driver generation"

        #     if len(node.input) > 0:
        #         producer = model.find_producer(node.input[0])
        #         init_tensor = model.get_initializer(node.input[0])
        #     else:
        #         producer = None
        #         init_tensor = None

        #     if producer is None:  # input dma?
        #         sdp_inst = getCustomOp(node)
        #         idma_name = sdp_inst.get_nodeattr("instance_name")
        #         df_model = ModelWrapper(sdp_inst.get_nodeattr("model"))
        #         assert df_model.graph.node[0].op_type == "IODMA"
        #         iodma_node = getCustomOp(df_model.graph.node[0])
        #         if iodma_node.get_nodeattr("burstMode") == "wrap":  # input weights dma?
        #             init_tensor = df_model.get_initializer(iodma_node.onnx_node.input[0])
        #             ext_weight_dma_cnt += 1
        #             w_dtype = df_model.get_tensor_datatype(iodma_node.onnx_node.input[0])
        #             init_external_tensor = to_external_tensor(init_tensor, w_dtype)
        #             np.save(weights_dir + "/" + idma_name + ".npy", init_external_tensor)
        #         idma_idx += 1

        return (model, False)


class MakePYNQDriverIODMA(Transformation):
    """Create PYNQ Python code to correctly interface the generated
    accelerator, including data packing/unpacking. Should be called
    after conversion to HLS layers, folding and the creation of
    dataflow partitions for correct operation.

    platform: one of ["zynq-iodma", "alveo"]

    Outcome if successful: sets the pynq_driver_dir attribute in the ONNX
    ModelProto's metadata_props field, with the created driver dir as the
    value. If any layers use runtime-writable parameters, those will be gathered
    under the runtime_weights/ subfolder of the pynq_driver_dir.
    """

    def __init__(self, platform, validation_datset=None):
        """Initialize MakePYNQDriverIODMA transformation.

        Args:
            platform: Target platform, one of ["zynq-iodma", "alveo"].
            validation_datset: Optional validation dataset. Defaults to None.
        """
        super().__init__()
        self.platform = platform
        self.validation_datset = validation_datset

    def apply(self, model):
        """Apply the MakePYNQDriverIODMA transformation.

        Creates a PYNQ Python driver package for interfacing with the generated
        accelerator, including data packing/unpacking and runtime weight handling.

        Args:
            model: The ONNX model to generate a driver for.

        Returns:
            Tuple of (modified model, False) indicating transformation applied.
        """
        # create a temporary folder for the generated driver
        pynq_driver_dir = make_build_dir(prefix="pynq_driver_")
        model.set_metadata_prop("pynq_driver_dir", pynq_driver_dir)

        # create the base FINN driver -- same for all accels
        driver_base_template = get_templates_folder() / "python_driver/driver_base.py"

        driver_base_py = pynq_driver_dir + "/driver_base.py"
        shutil.copy(driver_base_template, driver_base_py)
        # driver depends on qonnx and finn packages
        # extract individual source files and copy to driver folder
        qonnx_target_path = pynq_driver_dir + "/qonnx"
        finn_target_path = pynq_driver_dir + "/finn"
        os.makedirs(qonnx_target_path + "/core", exist_ok=True)
        os.makedirs(qonnx_target_path + "/util", exist_ok=True)
        os.makedirs(finn_target_path + "/util", exist_ok=True)
        qonnx_path = qonnx.__path__[0]
        finn_util_path = finn.util.__path__[0]
        files_to_copy = []
        files_to_copy.append(
            (qonnx_path + "/core/datatype.py", qonnx_target_path + "/core/datatype.py")
        )
        files_to_copy.append(
            (qonnx_path + "/core/__init__.py", qonnx_target_path + "/core/__init__.py")
        )
        files_to_copy.append((qonnx_path + "/util/basic.py", qonnx_target_path + "/util/basic.py"))
        files_to_copy.append(
            (qonnx_path + "/util/__init__.py", qonnx_target_path + "/util/__init__.py")
        )
        files_to_copy.append(
            (
                finn_util_path + "/data_packing.py",
                finn_target_path + "/util/data_packing.py",
            )
        )
        files_to_copy.append(
            (
                finn_util_path + "/__init__.py",
                finn_target_path + "/util/__init__.py",
            )
        )
        for src_file, target_file in files_to_copy:
            shutil.copy(src_file, target_file)

        driver_shapes: Dict = get_driver_shapes(model)

        # generate external weights npy files
        weights_dir = pynq_driver_dir + "/runtime_weights"

        os.makedirs(weights_dir)
        idma_idx = 0
        ext_weight_dma_cnt = 0

        for node in model.graph.node:
            assert (
                node.op_type == "StreamingDataflowPartition"
            ), "CreateDataflowPartition needs to be applied before driver generation"

            if len(node.input) > 0:
                producer = model.find_producer(node.input[0])
                init_tensor = model.get_initializer(node.input[0])
            else:
                producer = None
                init_tensor = None

            if producer is None:  # input dma?
                sdp_inst = getCustomOp(node)
                idma_name = sdp_inst.get_nodeattr("instance_name")
                df_model = ModelWrapper(sdp_inst.get_nodeattr("model"))
                assert df_model.graph.node[0].op_type == "IODMA_hls"
                iodma_node = getCustomOp(df_model.graph.node[0])
                # input weights dma?
                if iodma_node.get_nodeattr("burstMode") == "wrap":
                    init_tensor = df_model.get_initializer(iodma_node.onnx_node.input[0])
                    ext_weight_dma_cnt += 1
                    w_dtype = df_model.get_tensor_datatype(iodma_node.onnx_node.input[0])
                    init_external_tensor = to_external_tensor(init_tensor, w_dtype)
                    np.save(weights_dir + "/" + idma_name + ".npy", init_external_tensor)
                idma_idx += 1

        # fill in the driver template
        driver_py = pynq_driver_dir + "/driver.py"
        driver = template_driver.pynq_driver_template

        driver = driver.replace("$PLATFORM$", self.platform)
        driver = driver.replace("$INPUT_FINN_DATATYPE$", str(driver_shapes["idt"]).replace('"', ""))
        driver = driver.replace("$INPUT_SHAPE_NORMAL$", str(driver_shapes["ishape_normal"]))
        driver = driver.replace("$INPUT_SHAPE_FOLDED$", str(driver_shapes["ishape_folded"]))
        driver = driver.replace("$INPUT_SHAPE_PACKED$", str(driver_shapes["ishape_packed"]))
        driver = driver.replace(
            "$OUTPUT_FINN_DATATYPE$", str(driver_shapes["odt"]).replace('"', "")
        )
        driver = driver.replace("$OUTPUT_SHAPE_NORMAL$", str(driver_shapes["oshape_normal"]))
        driver = driver.replace("$OUTPUT_SHAPE_FOLDED$", str(driver_shapes["oshape_folded"]))
        driver = driver.replace("$OUTPUT_SHAPE_PACKED$", str(driver_shapes["oshape_packed"]))
        driver = driver.replace("$INPUT_DMA_NAME$", "%s" % str(driver_shapes["idma_names"]))
        driver = driver.replace("$OUTPUT_DMA_NAME$", "%s" % str(driver_shapes["odma_names"]))
        driver = driver.replace("$NUM_INPUTS$", str(len(driver_shapes["idma_names"])))
        driver = driver.replace("$NUM_OUTPUTS$", str(len(driver_shapes["odma_names"])))
        driver = driver.replace("$EXT_WEIGHT_NUM$", str(ext_weight_dma_cnt))

        with open(driver_py, "w") as f:
            f.write(driver)

        # add validate.py to run full top-1 test (only for suitable networks)
        validate_py = pynq_driver_dir + "/validate.py"
        validate_template = get_templates_folder() / "python_driver/validate.py"
        shutil.copy(validate_template, validate_py)

        # generate settings.json for generated driver
        if self.validation_datset is not None:
            settings = {
                "validation_dataset": self.validation_datset,
            }
            settingsfile = pynq_driver_dir + "/settings.json"
            with open(settingsfile, "w") as f:
                json.dump(settings, f, indent=2)

        # generate weight files for runtime-writable layers
        for sdp_ind, sdp_node in enumerate(model.graph.node):
            assert sdp_node.op_type == "StreamingDataflowPartition"
            # get dataflow model
            sdp_node = getCustomOp(sdp_node)
            dataflow_model_filename = sdp_node.get_nodeattr("model")
            dataflow_model = ModelWrapper(dataflow_model_filename)
            rt_layer_ind = 0
            for node in dataflow_model.graph.node:
                if node.op_type.startswith("MVAU") or node.op_type.startswith("Thresholding"):
                    node_inst = getCustomOp(node)
                    is_rt_weights = node_inst.get_nodeattr("runtime_writeable_weights")
                    if is_rt_weights == 1:
                        fcl_w = dataflow_model.get_initializer(node.input[1])
                        w_filename = weights_dir + "/%d_%d_%s.dat" % (
                            sdp_ind,
                            rt_layer_ind,
                            node.name,
                        )
                        node_inst.make_weight_file(fcl_w, "decoupled_runtime", w_filename)
                        rt_layer_ind += 1
                elif node.op_type == "StreamingDataflowPartition":
                    log.warning(
                        """Nested StreamingDataflowPartition are not supported
                    """
                    )
                else:
                    continue

        return (model, False)


class MakePYNQDriverInstrumentation(Transformation):
    """Create PYNQ Python driver with instrumentation for performance analysis.

    Generates a Python driver for PYNQ platforms that includes instrumentation
    capabilities for measuring performance metrics like throughput and latency.
    """

    def __init__(self, platform, clk_period_ns, live_fifo_sizing):
        """Initialize MakePYNQDriverInstrumentation transformation.

        Args:
            platform: Target platform identifier.
            clk_period_ns: Clock period in nanoseconds for performance calculations.
            live_fifo_sizing: Whether to include live FIFO sizing information.
        """
        super().__init__()
        self.platform = platform
        self.clk_period_ns = clk_period_ns
        self.live_fifo_sizing = live_fifo_sizing

    def apply(self, model: ModelWrapper):
        """Apply the MakePYNQDriverInstrumentation transformation.

        Generates a PYNQ driver with instrumentation capabilities for measuring
        performance metrics and analyzing FIFO sizing.

        Args:
            model: The ONNX model to generate an instrumented driver for.

        Returns:
            Tuple of (modified model, False) indicating transformation applied.
        """
        # TODO: support runtime-writable and external weights
        # TODO: support Alveo and Versal platforms

        # create a temporary folder for the generated driver
        pynq_driver_dir = make_build_dir(prefix="pynq_driver_")
        model.set_metadata_prop("pynq_driver_dir", pynq_driver_dir)

        # create (copy) the static instrumentation driver
        driver_template = get_templates_folder() / "python_driver/driver_instrumentation.py"
        if self.live_fifo_sizing:
            driver_py = pynq_driver_dir + "/driver_instrumentation.py"
        else:
            driver_py = pynq_driver_dir + "/driver.py"
        shutil.copy(driver_template, driver_py)

        # add-on driver for live fifosizing
        if self.live_fifo_sizing:
            driver_template = get_templates_folder() / "python_driver/driver_fifosizing.py"
            driver_py = pynq_driver_dir + "/driver.py"
            shutil.copy(driver_template, driver_py)

        # write default settings to driver config file
        settings = {
            "fclk_mhz": (1.0 / self.clk_period_ns) * 1e3,
        }
        if self.live_fifo_sizing:
            # export FIFO widths to the settings file as well
            # at this stage, the FIFOs are already wrapped in StreamingDataflowPartitions
            fifo_widths = {}
            for sdp_node in model.get_nodes_by_op_type("StreamingDataflowPartition"):
                sdp_node_inst = getCustomOp(sdp_node)
                # JSON doesn't support int keys
                sdp_id = str(sdp_node_inst.get_nodeattr("partition_id"))
                dataflow_model_filename = sdp_node_inst.get_nodeattr("model")
                kernel_model = ModelWrapper(dataflow_model_filename)
                for node in kernel_model.graph.node:
                    if node.op_type.startswith("StreamingFIFO"):
                        node_inst = getCustomOp(node)
                        fifo_widths[sdp_id] = node_inst.get_instream_width()
            settings["fifo_widths"] = fifo_widths
            # export original folding config to settings file,
            # so that the driver can generate a final cfg with live fifo sizes applied
            folding_path = model.get_metadata_prop("folding_config_before_lfs")
            if folding_path:
                with open(folding_path, "r") as f:
                    folding_cfg = json.load(f)
                settings["folding_config_before_lfs"] = folding_cfg

        settingsfile = pynq_driver_dir + "/settings.json"
        with open(settingsfile, "w") as f:
            json.dump(settings, f, indent=2)

        return (model, False)
